from sympy import *

r = Symbol("r")
el = Symbol("el")
az = Symbol("az")
a_n = Symbol("a_n")
a_e = Symbol("a_e")
a_z = Symbol("a_z")
c = Symbol("c")


def norm(arr):
    return sqrt(arr[0] * arr[0] + arr[1] * arr[1] + arr[2] * arr[2])


S = [r, el, az]
dt_0 = r / c
ant_nez = [a_n, a_e, a_z]
S_nez = [r * cos(az) * cos(el), r * sin(az) * cos(el), r * sin(el)]
delta_nez = [S_nez[0] - ant_nez[0], S_nez[1] - ant_nez[1], S_nez[2] - ant_nez[2]]
dt_ant = norm(delta_nez) / c

dt = dt_ant - dt_0

print(simplify(dt))

from math import sin, cos


def delay(r, el, az, a_n, a_e, a_z):
    c = 2.99793e8
    return (
        -r
        + (
            a_e ** 2
            - 2.0 * a_e * r * sin(az) * cos(el)
            + a_n ** 2
            - 2.0 * a_n * r * cos(az) * cos(el)
            + a_z ** 2
            - 2.0 * a_z * r * sin(el)
            + r ** 2
        )
        ** (1.0 / 2.0)
    ) / c


def assert_equal(x, y):
    dx = abs(x - y)
    if dx > 1e-5 * (abs(x) + abs(y)):
        print("FAIL %g != %g (%f)" % (x, y, dx))
    else:
        print("PASS %g == %g (%f)" % (x, y, dx))


r = 1e10


def to_rad(xx):
    return xx * 3.1415926 / 180.0


c = 2.99793e8
r = 1e9

assert_equal(delay(r, to_rad(90.0), to_rad(45), 0.0, 1.0, 0.0), 0.0)
assert_equal(delay(r, to_rad(45.0), to_rad(0), 0.0, 1.0, 0.0), 0.0)
assert_equal(delay(r, to_rad(0.0), to_rad(90), 0.0, 1.0, 0.0), -1.0 / c)
assert_equal(delay(r, to_rad(45.0), to_rad(45), 0.0, 1.0, 0.0), -1.6678e-9)



import math
import datetime
from .gps_time import GpsTime
from tart.util import utc


def get_utc_hours(timestamp):
    sec = (
        timestamp - datetime.datetime.combine(timestamp.date(), datetime.time.min)
    ).total_seconds()
    hours = sec / 3600.0
    return hours


def sign(x):
    return math.copysign(1, x)


def get_julian_date(timestamp):
    '''
        Check first that the timestamp is in UTC time
    '''
    utcoff = timestamp.utcoffset()
    if utcoff is not None:
        if utcoff.total_seconds() > 0:
            raise RuntimeError(f"Timestamp has utc offset {utcoff}")

    YY = timestamp.year
    MM = timestamp.month
    DD = timestamp.day
    UTcor = 0
    HR = timestamp.hour
    Min = timestamp.minute
    Sec = timestamp.second
    return (
        367 * YY
        - (7 * (YY + ((MM + 9) // 12)) // 4)
        + (275 * MM // 9)
        + DD
        + 1721013.5
        + UTcor // 24
        - 0.5 * sign((100 * YY) + MM - 190002.5)
        + 0.5
        + HR / 24.0
        + Min / (60.0 * 24.0)
        + Sec / (3600.0 * 24.0)
    )


def get_mjd(timestamp):
    return get_julian_date(timestamp) - 2400000.5


def JulianDay(utc_date):
    return get_julian_date(utc_date)



# -*- coding: utf-8 -*-
# This module takes an array of amplitudes of length X and returns an array of
# X independent random sources of required amplitudes and sample length n

import numpy as np
from tart.util import constants
from scipy import interpolate


class HorizontalSource(object):
    def __init__(self, r, azimuth, elevation):
        self.r = r
        self.azimuth = azimuth
        self.elevation = elevation


class SimulationSource(HorizontalSource):
    def __init__(self, r, amplitude, azimuth, elevation, sample_duration):
        super(SimulationSource, self).__init__(r, azimuth, elevation)
        self.omega = constants.L1_OMEGA
        self.duration = sample_duration
        self.amplitude = amplitude

        Fs = 16.368e6
        max_baseline = 100  # in m
        max_time = max_baseline / constants.V_LIGHT
        F_noise = 1.0e6 * np.power(np.pi, 3) / 2.9 / 3.7123 * 2.5

        from tart.simulation import butter_filter

        noisetime = np.arange(-max_time, self.duration + max_time, 1.0 / (F_noise))
        randnoise = np.random.uniform(-1.0, 1.0, len(noisetime))

        self.f = interpolate.InterpolatedUnivariateSpline(
            noisetime, self.amplitude * randnoise
        )

    def s_baseband(self, t):
        return self.f(t)

    def s(self, t):
        return self.s_baseband(t) * np.cos(self.omega * t)


if __name__ == "__main__":
    from . import spectrum
    import pylab
    from tart.util import angle
    import matplotlib.pyplot as plt

    crab = SimulationSource(
        r=1e10,
        amplitude=1.0,
        azimuth=angle.from_dms(0.0),
        elevation=angle.from_dms(45.0),
        sample_duration=1e-3,
    )  # 10 ms sample duration

    timebase = np.arange(0, 1e-5, 1e-6)
    for t in timebase:
        print("s(t)=%g s(t-tg)=%g" % (crab.s(t), crab.s(t - 1e-11)))

    sample_duration = 1e-4

    bb_timebase = np.arange(0, sample_duration, 1.0 / (16.384e6))
    crab_bbsignal = crab.s_baseband(bb_timebase)

    spectrum.plotSpectrum(crab_bbsignal, 16.384e6)
    pylab.show()

    from scipy import signal

    S_an = np.abs(np.fft.fft(signal.hilbert(crab_bbsignal)))
    freqs = np.fft.fftfreq(len(bb_timebase), 1.0 / 16.384e6)
    plt.figure()
    plt.plot(freqs, S_an)
    plt.show()

    radio_sampling_rate = 4.19e9
    print("Radio Sampling Rate %g" % radio_sampling_rate)

    timebase = np.arange(0, sample_duration, 1.0 / radio_sampling_rate)
    crab_signal = crab.s(timebase)

    spectrum.plotSpectrum(crab_signal, radio_sampling_rate)

    plt.figure()
    plt.plot(timebase, crab_signal, "-x", c="blue", label="full signal")
    plt.plot(bb_timebase, crab_bbsignal, c="red", label="baseband")
    plt.legend()
    plt.show()



'''
Usage: testbench.py [options] [test] [...]

Options:
  -h, --help       Show this message
  -v, --verbose    Verbose output
  -q, --quiet      Minimal output
  -f, --failfast   Stop on first failure
  -c, --catch      Catch control-C and display results
  -b, --buffer     Buffer stdout and stderr during test runs

Examples:
  testbench.py                               - run default set of tests
  testbench.py MyTestSuite                   - run suite 'MyTestSuite'
  testbench.py MyTestCase.testSomething      - run MyTestCase.testSomething
  testbench.py MyTestCase                    - run all 'test*' test methods
                                               in MyTestCase

  python testbench.py TestEphemeridesProxy.test_sp3_proxying
'''
import unittest
from tart.imaging.test.all import *
from tart.operation.test.all import *
from tart.simulation.test.all import *
from tart.util.test.all import *

if __name__ == '__main__':
    unittest.main()



import pickle

from tart.imaging import uvfitsgenerator
from tart.imaging import radio_source
from tart.imaging import location

from tart.simulation import antennas
from tart.util import skyloc
from tart.util import constants
from tart.util import angle

import numpy as np

# import pyfftw.interfaces.numpy_fft as fft
import numpy.fft as fft
import time

import os
import copy

import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

cc = np.concatenate


def get_max_ang(nw, num_bin):
    ret = np.degrees(num_bin / (4.0 * nw))
    return ret


class Synthesis_Imaging(object):
    def __init__(self, cal_vis_list, fixed_zenith=True):
        self.cal_vis_list = cal_vis_list
        self.fixed_zenith = fixed_zenith
        if self.fixed_zenith:
            self.phase_center = None
        else:
            vt = self.cal_vis_list[0]
            ra, dec = location.get_loc(vt.get_config()).horizontal_to_equatorial(
                vt.get_timestamp(), angle.from_dms(90), angle.from_dms(0)
            )
            self.phase_center = radio_source.CosmicSource(ra, dec, 1e10)
        self.grid_file = "grid.idx"
        self.grid_idx = None
        # print('debug:' , self.phase_center.to_horizontal(vt.config.get_loc(),vt.timestamp))

    def set_grid_file(self, fpath):
        self.grid_file = fpath

    def get_uvfits(self):
        gen = uvfitsgenerator.UVFitsGenerator(
            copy.deepcopy(self.cal_vis_list), self.phase_center
        )
        return gen

    def get_difmap_movie(self, base_index, frames):
        os.system("rm out.uvfits")
        fits_name = "out.uvfits"  # self.fname + ".uvfits"
        for i_frame in frames:
            v_list = copy.deepcopy(self.cal_vis_list)
            vis_indexes = base_index + i_frame
            vis_part = [v_list[ni] for ni in vis_indexes]

            # ra, dec = vis_part[0].config.get_loc().horizontal_to_equatorial(vis_part[0].timestamp, angle.from_dms(90.), angle.from_dms(0.))
            # self.phase_center = radio_source.CosmicSource(ra, dec)
            # print(ra, dec)

            uvgen = uvfitsgenerator.UVFitsGenerator(
                vis_part, self.phase_center
            )  # FIXME
            uvgen.write(fits_name)
            difcmd = get_difmap(fits_name, i_frame)
            f = open("difmap_cmds", "w")
            f.write(difcmd)
            f.close()
            os.system("difmap < difmap_cmds")
            os.system("rm out.uvfits")

    def get_uuvvwwvis_zenith(self):
        vis_l = []
        # for cal_vis in copy.deepcopy(self.cal_vis_list[:1]):
        for cal_vis in self.cal_vis_list[:1]:
            ant_p = np.array(cal_vis.get_config().get_antenna_positions())
            bls = cal_vis.get_baselines()
            pos_pairs = ant_p[np.array(bls)]
            uu_a, vv_a, ww_a = (
                pos_pairs[:, 0] - pos_pairs[:, 1]
            ).T / constants.L1_WAVELENGTH
            for bl in bls:
                ant_i, ant_j = bl
                vis_l.append(cal_vis.get_visibility(ant_i, ant_j))
        return uu_a, vv_a, ww_a, np.array(vis_l)

    def get_grid_idxs(self, uu_a, vv_a, num_bin, nw):
        try:
            if self.grid_idx is None:
                self.grid_idx = pickle.load(open(self.grid_file, "rb"))
                # print('finished loading ' + self.grid_file)
        except:
            print("generating...")
            uu_edges = np.linspace(-nw, nw, num_bin + 1)
            vv_edges = np.linspace(-nw, nw, num_bin + 1)
            grid_idx = []
            for uu, vv in zip(uu_a, vv_a):
                i = uu_edges.__lt__(uu).sum() - 1
                j = vv_edges.__lt__(vv).sum() - 1
                i2 = uu_edges.__lt__(-uu).sum() - 1
                j2 = vv_edges.__lt__(-vv).sum() - 1
                grid_idx.append([i, j, i2, j2])
            self.grid_idx = np.array(grid_idx)
            save_ptr = open(self.grid_file, "wb")
            pickle.dump(self.grid_idx, save_ptr, pickle.HIGHEST_PROTOCOL)
            save_ptr.close()
        return self.grid_idx

    def get_uvplane_zenith(
        self,
        num_bin=1600,
        nw=36,
    ):
        uu_a, vv_a, ww_a, vis_l = self.get_uuvvwwvis_zenith()
        arr = np.zeros((num_bin, num_bin), dtype=np.complex64)
        # place complex visibilities in the UV grid and prepare averaging by counting entries.
        grid_idxs = self.get_grid_idxs(uu_a, vv_a, num_bin, nw)
        count_arr = np.zeros((num_bin, num_bin), dtype=np.int16)
        # for k, v_l in enumerate(vis_l):
        #    i,j,i2,j2 = grid_idxs[k]
        #    count_arr[j, i] += 1
        #    count_arr[j2, i2] += 1
        if count_arr.max() > 1:
            # apply the masked array and divide by number of entries
            for k, v_l in enumerate(vis_l):
                i, j, i2, j2 = grid_idxs[k]
                arr[j, i] += v_l
                arr[j2, i2] += np.conjugate(v_l)
            n_arr = n_arr / (count_arr)
        else:
            arr[grid_idxs[:, 1], grid_idxs[:, 0]] = vis_l
            arr[grid_idxs[:, 3], grid_idxs[:, 2]] = np.conjugate(vis_l)
        n_arr = np.ma.masked_array(arr[:, :], count_arr.__lt__(1.0))
        return n_arr

    def get_uvplane(
        self, num_bin=1600, nw=36, grid_kernel_r_pixels=0.5, use_kernel=False
    ):

        for cal_vis in self.cal_vis_list:
            if self.fixed_zenith:
                uu_a, vv_a, ww_a = cal_vis.get_all_uvw()
                uu_a = uu_a / constants.L1_WAVELENGTH
                vv_a = vv_a / constants.L1_WAVELENGTH
                ww_a = ww_a / constants.L1_WAVELENGTH
                vis_l, bls = cal_vis.get_all_visibility()
            else:
                uu_l = []
                vv_l = []
                ww_l = []
                vis_l = []

                ts = cal_vis.get_timestamp()
                ra, dec = self.phase_center.radec(ts)
                c = cal_vis.get_config()
                ant_p = np.asarray(c.get_antenna_positions())
                loc = location.get_loc(c)
                bls = cal_vis.get_baselines()

                for bl in bls:
                    ant_i, ant_j = bl
                    a0 = antennas.Antenna(loc, ant_p[ant_i])
                    a1 = antennas.Antenna(loc, ant_p[ant_j])
                    uu, vv, ww = antennas.get_UVW(a0, a1, ts, ra, dec)
                    uu_l.append(uu / constants.L1_WAVELENGTH)
                    vv_l.append(vv / constants.L1_WAVELENGTH)
                    ww_l.append(ww / constants.L1_WAVELENGTH)
                    vis_l.append(cal_vis.get_visibility(ant_i, ant_j))
                uu_a = np.asarray(uu_l)
                vv_a = np.asarray(vv_l)
                ww_a = np.asarray(ww_l)
                vis_l = np.asarray(vis_l)

        uu_edges = np.linspace(-nw, nw, num_bin + 1)
        vv_edges = np.linspace(-nw, nw, num_bin + 1)

        if use_kernel == False:

            n_arr = np.zeros((num_bin, num_bin), dtype=np.complex64)
            uu_comb = np.concatenate((uu_a, -uu_a))
            vv_comb = np.concatenate((vv_a, -vv_a))
            all_v = np.concatenate((vis_l, np.conjugate(vis_l)))
            h_real, _, _ = np.histogram2d(
                vv_comb, uu_comb, weights=all_v.real, bins=[vv_edges, uu_edges]
            )
            h_imag, _, _ = np.histogram2d(
                vv_comb, uu_comb, weights=all_v.imag, bins=[vv_edges, uu_edges]
            )
            num_entries, _, _ = np.histogram2d(
                vv_comb, uu_comb, bins=[vv_edges, uu_edges]
            )
            n_arr[:, :] = h_real + (1j * h_imag)
            pos = np.where(num_entries.__gt__(1))
            n_arr[pos] /= num_entries[pos]

        else:
            pixels_per_wavelength = num_bin / (nw * 2.0)
            halfbin = float(nw) / (num_bin)
            mid_points_uv = np.mgrid[
                -nw + halfbin : nw - halfbin : num_bin * 1j,
                -nw + halfbin : nw - halfbin : num_bin * 1j,
            ]
            n_arr = np.zeros((num_bin, num_bin), dtype=complex)

            r_noise_wavelengths = 0.1
            grid_kernel_r_wavelength = (
                r_noise_wavelengths + grid_kernel_r_pixels / pixels_per_wavelength
            )
            offset_px = np.ceil(grid_kernel_r_wavelength * pixels_per_wavelength)
            offsets = np.arange(-offset_px, offset_px + 1)

            vis_max_abs = np.max(np.abs(vis_l))
            for uu, vv, v_l in zip(uu_a, vv_a, vis_l):
                i = uu_edges.__lt__(uu).sum() - 1
                j = vv_edges.__lt__(vv).sum() - 1
                # print('u', mid_points_uv[0][i-1,0], mid_points_uv[0][i,0], mid_points_uv[0][i+1,0], uu)
                # print('v', mid_points_uv[1][0,j], vv)
                for i_offset in offsets:
                    for j_offset in offsets:
                        r = np.sqrt(
                            np.power(uu - mid_points_uv[0][i + i_offset, 0], 2)
                            + np.power(vv - mid_points_uv[1][0, j + j_offset], 2)
                        )
                        # print('u', uu- mid_points_uv[0][i+i_offset,0], r , grid_kernel_r_pixels/pixels_per_wavelength)
                        # print('v', vv- mid_points_uv[1][0,j+j_offset], r , grid_kernel_r_pixels/pixels_per_wavelength)
                        n_arr[j + j_offset, i + i_offset] += v_l * np.exp(
                            -(r ** 2.0 / (grid_kernel_r_wavelength) ** 2.0)
                        )
                        # print(i,j, i_offset, j_offset, r, v_l * np.exp(-(r**2. / (grid_kernel_r_pixels/pixels_per_wavelength)**2.)))
                i = uu_edges.__lt__(-uu).sum() - 1
                j = vv_edges.__lt__(-vv).sum() - 1
                for i_offset in offsets:
                    for j_offset in offsets:
                        r = np.sqrt(
                            np.power(-uu - mid_points_uv[0][i + i_offset, 0], 2)
                            + np.power(-vv - mid_points_uv[1][0, j + j_offset], 2)
                        )
                        n_arr[j + j_offset, i + i_offset] += np.conjugate(v_l) * np.exp(
                            -(r ** 2 / (grid_kernel_r_wavelength) ** 2)
                        )
                # apply the masked array and divide by number of entries
                mask = np.abs(n_arr).__gt__(0.0)
                # mask = np.abs(n_arr).__gt__(vis_max_abs)
                n_arr[mask] = n_arr[mask] / np.abs(n_arr[mask])
        return (n_arr, uu_edges, vv_edges)

    def get_ift_simp(self, nw=30, num_bin=2 ** 7):
        uv_plane = self.get_uvplane_zenith(num_bin=num_bin, nw=nw)
        ift = np.fft.fftshift(fft.ifft2(np.fft.ifftshift(uv_plane)))
        maxang = get_max_ang(nw, num_bin)
        extent = [maxang, -maxang, -maxang, maxang]
        return [ift, extent]

    def get_ift(self, nw=30, num_bin=2 ** 7, use_kernel=False):
        uv_plane, uu_edges, vv_edges = self.get_uvplane(
            num_bin=num_bin, nw=nw, use_kernel=use_kernel
        )
        ift = np.fft.fftshift(fft.ifft2(np.fft.ifftshift(uv_plane)))
        maxang = get_max_ang(nw, num_bin)
        extent = [maxang, -maxang, -maxang, maxang]
        return [ift, extent]

    def get_beam(self, nw=30, num_bin=2 ** 7, use_kernel=False):
        uv_plane, uu_edges, vv_edges = self.get_uvplane(
            num_bin=num_bin, nw=nw, use_kernel=use_kernel
        )
        ift = np.fft.ifftshift(fft.ifft2(np.fft.ifftshift(np.abs(uv_plane).__gt__(0))))
        return ift  # /np.sum(ret)

    def get_image(self, CAL_IFT, CAL_EXTENT):
        abs_ift = np.abs(CAL_IFT)
        ift_std = np.std(abs_ift)
        ift_scaled = abs_ift / ift_std

        plt.figure(figsize=(8, 6))
        plt.imshow(ift_scaled, extent=CAL_EXTENT)
        plt.colorbar()
        ift_median = np.median(abs_ift)
        ts = self.cal_vis_list[0].get_timestamp()
        plt.title(
            ts.strftime("%d-%m-%y %H:%M:%S")
            + " std %1.3e median: %1.3e" % (ift_std, ift_median)
        )
        plt.grid()
        plt.xlim(63, -63)
        plt.ylim(-63, 63)
        plt.xlabel("East-West")
        plt.ylabel("North-South")
        plt.tight_layout()



import numpy as np

def unwrap(p):
    num_wrap = int(np.abs(p)/(2*np.pi))
    if p < 0:
        ret = p + (num_wrap+1)*2.*np.pi
    elif p >= 2.*np.pi:
        ret = p - num_wrap*2.*np.pi
    else:
        ret = p
    return ret

def compare_phases(p1, p2):
    up1 = unwrap(p1)
    up2 = unwrap(p2)
    diff = np.abs(up1 - up2)
    if diff > np.pi:
        ret = 2.*np.pi-diff
    else:
        ret = diff
    return ret




#!/usr/bin/python
# Simulate the known radio sources during an entire day's observing
# Calculate visibilities and compare them to the observed visibilities
#
# Copyright (c) Tim Molteno 2013-2019. tim@elec.ac.nz
#
import traceback
import argparse
import math
import utc

import numpy as np
from multiprocessing import Pool


import datetime

import matplotlib.pyplot as plt
from tart.util import angle

import forward_map

from .visibility import *
from . import location
from ..operation import observation


def process_file(
    dirname, fname, orientation_degrees, spacing, sample_duration, simulate_vis=True
):
    try:
        fpath = os.path.join(dirname, fname)
        nobs = observation.Observation_Load(fpath)
        date = nobs.timestamp.date()
        sec = (
            nobs.timestamp - datetime.datetime.combine(date, datetime.time.min)
        ).total_seconds()

        if simulate_vis == True:
            utc_date = nobs.timestamp.replace(tzinfo=utc.UTC)
            antenna_locations = antenna_locations = nobs.ant_positions

            sky, baselines, visibility = forward_map.forward_map(
                utc_date,
                location.Dunedin,
                horizon=5.0,
                antenna_locations=antenna_locations,
                sample_duration=sample_duration,
            )  # Remove things below an elevation of five degrees
        else:
            sky = False
            baselines = False
            visibility = [False]

        hours = sec / 3600.0
        v = Visibility(nobs)
        return date, hours, v, visibility[0]
    except Exception as inst:
        print(type(inst))  # the exception instance
        print(inst.args)  # arguments stored in .args
        print(inst)
        print(sample_duration)
        print(traceback.format_exc())
        return None, None, None, None
    except KeyboardInterrupt:
        print(inst)
        return None, None, None, None


def get_correlation(
    dirname, orientation_degrees, spacing, sample_duration, simulate_vis=True
):
    files = sorted(os.listdir(dirname))
    vis = []
    simvis = []
    times = []

    if True:
        p = Pool()

        resultList = []

        try:
            for fname in files:
                if "data.pkl" in fname:
                    resultList.append(
                        p.apply_async(
                            process_file,
                            (
                                dirname,
                                fname,
                                orientation_degrees,
                                spacing,
                                sample_duration,
                                simulate_vis,
                            ),
                        )
                    )

            p.close
            p.join

            for sv_thread in resultList:
                [date, hours, v, sv] = sv_thread.get()  # timeout=2000
                if date != None:
                    times.append(hours)
                    vis.append(v)
                    simvis.append(sv)
                    # print("%g %g %g" % (times[-1], vis[-1], simvis[-1]))
            p.terminate()
        except KeyboardInterrupt:
            print("control-c pressed")
            p.terminate()

    else:
        for fname in files:
            if "data.pkl" in fname:
                [date, hours, v, sv] = process_file(
                    dirname,
                    fname,
                    orientation_degrees,
                    spacing,
                    sample_duration,
                    simulate_vis,
                )

                times.append(hours)
                vis.append(v)
                simvis.append(sv)

                print("%g %s %s" % (times[-1], v.v, simvis))

    # vis = np.array(vis)
    simvis = np.array(simvis)
    return date, times, vis, simvis


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Simulate the Sky and calculate simulated visibilities."
    )
    parser.add_argument(
        "--data-dir", required=True, help="The filesystem path for the telescope data."
    )
    parser.add_argument(
        "--orientation-angle",
        required=True,
        type=float,
        default=-22.0,
        help="The orientation of the array (degrees).",
    )
    parser.add_argument(
        "--optimize",
        action="store_true",
        help="Optimize the correlation between the simulated and measured visibilities.",
    )
    parser.add_argument(
        "--spacing", type=float, default=0.5, help="The spacing between antennas."
    )
    parser.add_argument(
        "--sample-duration",
        type=float,
        default=1e-4,
        help="The length of time to simulate data for.",
    )
    args = parser.parse_args()
    dirname = args.data_dir
    sample_duration = args.sample_duration
    orientation_degrees = args.orientation_angle
    spacing = args.spacing

    if args.optimize:
        for theta in np.arange(
            orientation_degrees - 10.0, orientation_degrees + 10.0, 1.0
        ):
            for r in np.arange(spacing - 0.03, spacing + 0.03, 0.01):

                date, times, vis, simvis = get_correlation(
                    dirname, theta, r, sample_duration
                )
                corr = sum(vis * simvis)
                print("Correlation @%f %f %f" % (theta, r, corr))
    else:
        date, times, vis, simvis = get_correlation(
            dirname, orientation_degrees, spacing, sample_duration
        )
        corr = sum(vis * simvis)
        print("Correlation @%f %f %f" % (orientation_degrees, spacing, corr))

        plt.subplot(2, 1, 1)
        title = "Visibility Data from %s (th=%f, dx=%f)" % (
            str(date),
            orientation_degrees,
            spacing,
        )
        plt.title(title)
        plt.plot(times, vis, ".")
        plt.ylabel("Measured V")
        plt.xlabel("Hours since UTC ")
        plt.subplot(2, 1, 2)
        plt.plot(times, simvis, ".")
        plt.ylabel("Simulated V")
        plt.xlabel("Hours since UTC ")
        plt.savefig(
            os.path.join(
                dirname,
                "simulated_visibility_%f_%f.pdf" % (orientation_degrees, spacing),
            )
        )
        plt.show()



# -*- coding: utf-8 -*-
#
# Copyright (c) Tim Molteno 2013. tim@elec.ac.nz
import numpy as np
import math
import time
import scipy.signal


from tart.operation import observation
from tart.simulation import antennas
from tart.simulation import butter_filter


class Radio(object):
    def sampled_signal(ant_signal):
        raise RuntimeError(
            "Radio must implement a function sampled_signal() that returns the sampled signal."
        )


class Max2769B(Radio):
    """
    MAX2967B in preconfigured mode 2 has an IF center 4.092 MHz and a bandwidth of 2.5 MHz. The GPS L1 frequency is 1575.42 MHz
    The LO Frequency Range is from 1550 - 1610
    The reference division ratio is 16
    The main division ratio is 1536
    16.368 / 16 = 1.023
    1.023 * 1536 = 1571.328 - the local oscillator frequency
    1571.328 + 4.092  = 1575.42
    Thus mixing incoming L1 signal with the LO will give an IF of 4.092 MHz as required

    Note that Low Side Local Oscillator injection is used so image at 1567.236 MHz must be rejected

    Excerpts from the data sheet are as follows;
    .... including a dual-input LNA and mixer, FOLLOWED BY AN IMAGE REJECTED FILTER, PGA, VCO,
    fractional-N frequency synthesizer, crystal oscillator, and a multibit ADC.
    ... The MAX2769B completely eliminates the need for exter-nal IF filters by implementing on-chip monolithic filters
    ... Note 3: The LNA output connects to the mixer input without a SAW filter between them. (Case in default mode)
    The output of the LNA and the input of the mixer are brought off-chip to facilitate the use of a SAW filter.
    """

    def __init__(
        self,
        noise_level,
        n_samples=2 ** 15,
        ref_freq=16.368e6,
        freq_mult=256,
        ref_div=16,
        main_div=1536,
        bandwidth=2.5e6,
        order=5,
    ):
        self.sample_duration = n_samples * 1.0 / ref_freq  # sample_duration
        self.n_samples = n_samples
        self.ref_freq = ref_freq
        self.freq_mult = freq_mult
        self.ref_div = ref_div
        self.main_div = main_div
        self.bandwidth = bandwidth
        self.order = order
        self.noise_level = noise_level
        self.sampling_rate = self.ref_freq * self.freq_mult
        self.baseband_timebase = np.arange(0, self.n_samples) * 1.0 / self.ref_freq
        self.int_freq = self.ref_freq / self.ref_div * 4
        # print(self.int_freq)

    def sampled_signal(self, ant_signal, ant_index, sample_duration, debug=True):
        # t = np.linspace(0, self.sample_duration, len(ant_signal))
        # t = np.arange(0, self.sample_duration, self.sample_duration/len(ant_signal))
        t = np.arange(0, sample_duration, sample_duration / len(ant_signal))

        # Produce the LO signal - massively oversampled so approximates continuous time
        lo_freq = self.ref_freq / self.ref_div * self.main_div  # LO frequency
        lo_omega = lo_freq * 2 * np.pi  # LO in radians/s
        lo = np.sin(lo_omega * t)  # Produce sinusoidal LO signal of reqd length

        if debug:
            print("LO Frequency: %f" % lo_freq)
        if debug:
            print("Intermediate Frequency: %f" % self.int_freq)

        # Mix the incoming signal with LO to generate the IF
        if_sig = lo * ant_signal
        # if debug: print('if_sig', if_sigk         json)

        # Need anti-aliasing filter BEFORE downsampling. See http://en.wikipedia.org/wiki/Downsampling
        # This is a low pass filter with a cutoff of ref_freq / 2.
        cutoff_freq = self.int_freq * 1.5
        samp_rate = float(len(ant_signal)) / sample_duration
        # print(samp_rate)
        # The Nyquist rate of the signal.
        nyq_rate = samp_rate / 2.0
        width = 2e6 / nyq_rate

        # design filter
        b, a = scipy.signal.butter(3, cutoff_freq / nyq_rate)
        alias_sig = scipy.signal.lfilter(b, a, if_sig)

        # Decimate the signal data by a factor of freq_mult
        samp_sig = alias_sig[:: self.freq_mult]
        if debug:
            print("samp_sig", samp_sig)

        if self.noise_level[ant_index] > 0.0:
            noise = np.random.normal(0, self.noise_level[ant_index], samp_sig.size)
            samp_sig = samp_sig + noise

        # Implement 5th order Butterworth filter as used by the max2769B
        filt_sig = butter_filter.butter_bandpass_filter(
            samp_sig,
            self.int_freq - self.bandwidth / 2,
            self.int_freq + self.bandwidth / 2,
            self.ref_freq,
            self.order,
        )
        if debug:
            print("filt_sig", filt_sig)

        # Convert the filtered and sampled signal to one bit NRZ binary format
        filt_sig1 = np.sign(filt_sig)  # -1 if negative, 0 if 0, +1 if positive
        filt_sig1[filt_sig == 0] = 1  # Replace 0 with 1 - true NRZ
        if debug:
            print("filt_sig1", filt_sig1)

        return filt_sig1

    def get_full_obs(self, ant_sigs, utc_date, config, timebase):
        num_radio_samples = len(timebase) // self.freq_mult
        num_ant = len(ant_sigs)
        sampled_signals = np.zeros((num_ant, num_radio_samples))
        for i in range(num_ant):
            sampled_signals[i, :] = self.sampled_signal(
                ant_sigs[i], i, self.sample_duration
            )
        sampled_signals = np.asarray(
            (sampled_signals + 1) / 2, dtype=bool
        )  # turn into boolean array
        obs = observation.Observation(utc_date, config=config, data=sampled_signals)
        return obs

    def get_simplified_obs(self, baseband_signals, utc_date, config, seed=None):
        np.random.seed(seed=seed)
        s_signals = []
        if_sig = baseband_signals * np.exp(
            -2.0j * np.pi * self.int_freq * self.baseband_timebase
        )
        num_ant = len(baseband_signals)
        for i in range(num_ant):
            if self.noise_level[i] > 0.0:
                noise = np.random.normal(0.0, self.noise_level[i], len(if_sig[i]))
                if_sig[i] = if_sig[i] + noise

        for ant_num in range(num_ant):
            filt_sig1 = butter_filter.butter_bandpass_filter(
                if_sig[ant_num],
                self.int_freq - self.bandwidth / 2.0,
                self.int_freq + self.bandwidth / 2.0,
                self.ref_freq,
                self.order,
            )
            s_signals.append(filt_sig1)

        s_signals = np.array(s_signals).real
        sampled_signals = np.sign(s_signals)  # -1 if negative, 0 if 0, +1 if positive
        sampled_signals[s_signals == 0.0] = 1.0  # Replace 0 with 1 - true NRZ
        sampled_signals = np.asarray(
            (sampled_signals + 1) / 2, dtype=bool
        )  # turn into boolean array
        obs = observation.Observation(utc_date, config=config, data=sampled_signals)
        return obs


if __name__ == "__main__":
    import numpy as np
    import datetime
    import matplotlib.pyplot as plt

    from tart.operation import settings
    from tart.simulation import simulation_source
    from tart.simulation import antennas
    from tart.simulation import spectrum
    from tart.imaging import antenna_model
    from tart.util import angle

    from tart.simulation.radio import *

    config = settings.Settings("../test/test_telescope_config.json")
    num_ant = config.get_num_antenna()

    noiselvls = 0.1 * np.ones(num_ant)
    rad = Max2769B(n_samples=2 ** 14, noise_level=noiselvls)
    sources = [
        simulation_source.SimulationSource(
            amplitude=1.0,
            azimuth=angle.from_dms(0.0),
            elevation=angle.from_dms(90.0),
            sample_duration=rad.sample_duration,
        )
    ]
    ants = [antennas.Antenna(config.get_loc(), pos) for pos in config.ant_positions]
    ant_models = [antenna_model.GpsPatchAntenna() for i in range(num_ant)]
    utc_date = datetime.datetime.utcnow()

    plt.figure()
    ant_sigs = antennas.antennas_signal(ants, ant_models, sources, rad.timebase)
    rad_sig_full = rad.sampled_signal(ant_sigs[0, :], 0, rad.sample_duration)
    obs_full = rad.get_full_obs(ant_sigs, utc_date, config)

    ant_sigs_simp = antennas.antennas_simplified_signal(
        ants, ant_models, sources, rad.baseband_timebase, rad.int_freq
    )
    obs_simp = rad.get_simplified_obs(ant_sigs_simp, utc_date, config)

    from matplotlib import mlab

    power, freq = mlab.psd(rad_sig_full, Fs=rad.ref_freq, NFFT=8192)
    plt.plot(freq / 1e6, 10.0 * np.log10(power), label="full_before_obs_obj", c="blue")
    power, freq = mlab.psd(obs_full.get_antenna(0), Fs=rad.ref_freq, NFFT=8192)
    plt.plot(freq / 1e6, 10.0 * np.log10(power), label="full", c="cyan")
    power, freq = mlab.psd(obs_simp.get_antenna(0), Fs=rad.ref_freq, NFFT=8192)
    plt.plot(freq / 1e6, 10.0 * np.log10(power), label="simp", c="red")

    # freqs, spec_full_before_obs = spectrum.plotSpectrum(rad_sig_full, rad.ref_freq, label='full_before_obs_obj', c='blue')
    # freqs, spec_full = spectrum.plotSpectrum(obs_full.get_antenna(0), rad.ref_freq, label='full', c='cyan')
    # freqs, spec_simp = spectrum.plotSpectrum(obs_simp.get_antenna(0), rad.ref_freq, label='simp', c='red')
    plt.legend()
    plt.show()

    # assertTrue((spec_full_before_obs == spec_full).all(), True)

    # plt.figure()
    # plt.plot(freqs, (spec_simp-spec_full)/spec_full)

    print(len(obs_full.get_antenna(0)), obs_full.get_antenna(0).mean())
    print(len(obs_simp.get_antenna(0)), obs_simp.get_antenna(0).mean())



import unittest
import numpy as np
from tart.imaging.ephemerides_proxy import *


class TestEphemeridesProxy(unittest.TestCase):
    def setUp(self):
        self.ep = EphemeridesProxy.Instance()

    def test_proxying(self):
        t = utc.utc_datetime(2013, 9, 21, 0, 59, 3)
        sv = 21
        for i in range(0, 100):
            t = t + datetime.timedelta(seconds=1.0)
            pos = self.ep.get_sv_position(t, sv)
            pos_remote = self.ep.get_remote_position(t, sv)
            diff = np.array(pos) - np.array(pos_remote)
            dr = np.sqrt(diff.dot(diff))
            if dr > 3.0:
                print("test_proxying %i %s %f %s %s" % (i, t, dr, pos, pos_remote))
            self.assertLess(dr, 3.0)  # Maximum difference of 3 meters

    # TODO sp3 orbit interpolation from the precise sp3 positions
    # Eg at ftp://cddis.gsfc.nasa.gov/pub/gps/products/1172/
    # or ftp://nfs.kasi.re.kr/glonass/products/1355/
    # And compare against predicted files. There is a BUG in
    # the ephemeris file.

    # def test_all_positions(self):
    # t = utc.utc_datetime(2012, 10, 31, 0, 2, 2)
    # for i in range(0,10):
    # t = t + datetime.timedelta(seconds=100.0)
    # pos = self.ep.get_sv_positions(t)

    def test_cache_jump(self):
        t1 = utc.utc_datetime(2013, 9, 21, 0, 59, 3)
        t = utc.utc_datetime(2013, 9, 21, 0, 59, 3)
        sv = 21
        p1 = self.ep.get_sv_position(t1, sv)
        for i in range(0, 100):
            t = t + datetime.timedelta(seconds=1.0)
            p2 = self.ep.get_sv_position(t, sv)
            diff = np.array(p2) - np.array(p1)
            dr = np.sqrt(diff.dot(diff))
            if dr > 1000:
                print("test_cache_jump %i %s %f" % (i, p2, dr))
            self.assertLess(dr, 3000.0 * (i + 1))

    def test_sp3_proxying(self):
        t = utc.utc_datetime(2013, 9, 21, 0, 59, 3)
        sv = 21
        for i in range(0, 100):
            t = t + datetime.timedelta(seconds=1.0)
            pos = self.ep.get_sv_position_sp3(t, sv)
            pos_remote = self.ep.get_remote_position(t, sv)
            diff = np.array(pos) - np.array(pos_remote)
            dr = np.sqrt(diff.dot(diff))
            if dr > 1.0:
                print("test_sp3_proxying %i %s %f %s %s" % (i, t, dr, pos, pos_remote))
            self.assertLess(dr, 1.0)  # Maximum difference of 1 meters



"""
    Antenna Models for TART
    
    Author: Tim Molteno 2014-2019 tim@elec.ac.nz
    Copyright GPLv3
"""
import numpy as np
import healpy as hp

from tart.imaging import location
from tart.util import angle
from tart.util.db import db_connect, db_remove_antenna_measurements, db_insert_row


class AntennaModel(object):
    """Base class for all Antenna models."""

    def get_gain(self, el, az):
        """Dummy method. All base classes should override this"""
        raise RuntimeError("You should be using a subclass of AntennaModel")

    def get_gain_equatorial(self, loc, utc_time, ra, decl):
        """Return the gain from equatorial co-ordinates"""
        el, az = loc.equatorial_to_horizontal(utc_time, ra, decl)
        return self.get_gain(el, az)


class IdealHemisphericalAntenna(AntennaModel):
    """Ideal Hemispherical Antenna."""

    """Return the gain """

    def get_gain(self, el, az):
        if el.to_degrees() < 0.0:
            return 0.0
        if el.to_degrees() > 90:
            return 0.0
        return 1.0


class GpsPatchAntenna(AntennaModel):
    """A more realistic model of a GPS patch antenna."""

    def get_gain(self, el, az):
        if el.to_degrees() < 5.0:
            return 0.0
        if el.to_degrees() > 90:
            return 0.0
        return 1.0


def hp_interpolator(map_, el, az, n_pix=4):
    NSide = hp.pixelfunc.get_nside(map_)
    direction = np.array([np.pi / 2.0 - el.to_rad(), az.to_rad()])
    steplength = hp.pixelfunc.max_pixrad(NSide)
    for i, r in enumerate(np.arange(steplength, np.pi, steplength)):
        pixels = np.array(
            hp.query_disc(NSide, hp.ang2vec(direction[0], direction[1]), r)
        )
        filled = np.where(map_[pixels] > -1.0)[0]
        l = len(filled)
        if l >= n_pix:
            # print(i, l)
            filled_pixel = pixels[filled]
            filled_pixel_directions = hp.pix2vec(NSide, filled_pixel)
            angular_distance = hp.rotator.angdist(direction, filled_pixel_directions)
            if angular_distance.min() == 0.0:  # do we really want this?
                return map_[filled_pixel[angular_distance.argmin()]]
            return np.average(
                map_[filled_pixel], weights=np.power(1.0 / angular_distance, 2)
            )


# def hp_spheric_harmonics(hp_alm, el, az, mmax=-1):
# from scipy.special import sph_harm
# if mmax==-1:
# lmax = hp.Alm.getlmax(len(hp_alm))
# mmax = lmax
# else:
# lmax = hp.Alm.getlmax(len(hp_alm),mmax=mmax)

# print(lmax, mmax)
# out = 0.
# azimutal_angle = az.to_rad()
# polar_angle = el.to_rad()-np.pi/2.
# for l in range(lmax+1):
# out += np.real(hp_alm[hp.Alm.getidx(lmax,l,0)] * sph_harm(0, l, azimutal_angle, polar_angle))
# for m in range(1, 1+min(l , mmax)):
# Ylm = sph_harm(m, l, azimutal_angle, polar_angle)
# out += np.real(hp_alm[hp.Alm.getidx(lmax,l,m)] * Ylm)
# out += np.real(hp_alm[hp.Alm.getidx(lmax,l,m)] * Ylm.conj()/np.power(-1.,m))
# return out


def gen_interpolation_map(points, values, antenna_num, nside_exp):
    nside = np.power(2, nside_exp)
    npix = hp.nside2npix(nside)
    hp_map_avg = np.ones(npix) * hp.UNSEEN
    theta = np.array([90.0 - i[0] for i in points]) * np.pi / 180.0
    phi = np.array([i[1] for i in points]) * np.pi / 180.0

    pix = hp.pixelfunc.ang2pix(nside, theta, phi)

    values = np.array(values)

    dic = {}
    for i, p in enumerate(pix):
        if dic.get(p) is None:
            dic[p] = []
        dic[p].append(values[i])

    pixel_dict = {}
    for p in dic:
        pixel_dict[p] = np.array(dic[p]).mean()
        hp_map_avg[p] = pixel_dict[p]

    map_mean = np.array([pixel_dict[key] for key in pixel_dict]).mean()

    for p in dic:
        # pixel_dict[p] = (pixel_dict[p]-map_mean)*(1.*len(hp_map_avg))/(4.*np.pi)
        hp_map_avg[p] = pixel_dict[p]

    return (hp_map_avg, pixel_dict)


class EmpiricalAntenna(AntennaModel):
    """An antenna model that has a radiation pattern built up
    from measured GPS signal correlation amplitudes. This
    incorporates an horizon model as well.

    The model can be serialized into JSON form using the to_json()
    method. And reconstructed from JSON using the from_json()
    """

    def __init__(self, antenna_num):
        self.sv = []  # A list of sv numbers - essentially an identifier
        self.distances = []  # A list of sv distances - essentially an identifier
        self.points = []  # The data is stored as a list of el, az.
        self.values = []  # The data is stored as a list of amplitudes.
        self.recalculate = True
        self.antenna_num = antenna_num
        self.interpolation_cache = []

    def add_measurement(self, el, az, amplitude, sv=-1, distance=-1):
        self.points.append([el.to_degrees(), az.to_degrees()])
        self.values.append(amplitude)
        self.sv.append(sv)
        self.distances.append(distance)
        self.recalculate = True

    def get_gain(
        self,
        el,
        az,
        n_pix=4,
        nside_exp_grid=10,
        nside_exp_syn=10,
        lmax=4,
        mmax=4,
        interpolate="default",
    ):
        if self.recalculate:
            self.recalculate = False
            self.values = np.array(self.values)
            self.distances = np.array(self.distances)
            self.i_map, self.pixel_dict = gen_interpolation_map(
                self.points, self.values, self.antenna_num, nside_exp_grid
            )
            if interpolate == "increasing_neighborhood":
                self.interp_gain = lambda el, az: hp_interpolator(
                    self.i_map, el, az, n_pix
                )

                # self.interp_gain(el, az)
                np.power(2, nside_exp_syn)
                self.norm_map = self.i_map
            else:
                alms = hp.sphtfunc.map2alm(
                    hp.ma(self.i_map), lmax=lmax, mmax=mmax, iter=10, pol=False
                )
                print(hp.Alm.getlmax(len(alms), mmax=mmax))
                norm_map = hp.sphtfunc.alm2map(
                    alms,
                    np.power(2, nside_exp_syn),
                    lmax=lmax,
                    mmax=mmax,
                    pixwin=False,
                    pol=False,
                )
                norm_map_min = norm_map.min()
                norm_map = norm_map - norm_map_min
                norm_map_max = norm_map.max()
                norm_map = norm_map / norm_map_max
                self.norm_map = norm_map
                # self.interp_gain = lambda el, az: (hp_spheric_harmonics(alms, el, az, mmax=mmax)-norm_map_min)/norm_map_max
                self.interp_gain = lambda el, az: self.norm_map[
                    hp.pixelfunc.ang2pix(
                        np.power(2, nside_exp_syn), np.pi / 2 - el.to_rad(), az.to_rad()
                    )
                ]
        return self.interp_gain(el, az)

    def to_json(self, filename):
        import json

        ret = json.dumps(
            {
                "antenna_num": self.antenna_num,
                "sv": self.sv,
                "distances": self.distances,
                "points": self.points,
                "values": self.values,
            },
            indent=4,
            separators=(",", ": "),
        )
        f = open(filename, "w")
        f.write(ret)
        f.close()

    def to_db(self, utc_date, db_file=None, table="gps_signals"):
        conn, sql = db_connect(db_file, table=table)

        db_remove_antenna_measurements(
            conn, sql, utc_date, self.antenna_num, table=table
        )
        cursor = conn.cursor()
        for p, v, sv, distance in zip(
            self.points, self.values, self.sv, self.distances
        ):
            el, az = p
            correlation = v
            db_insert_row(
                cursor,
                sql,
                utc_date,
                self.antenna_num,
                sv,
                el,
                az,
                correlation,
                distance,
                table=table,
            )

        conn.commit()
        conn.close()

    @classmethod
    def from_json(self, filename):
        import json

        f = open(filename, "r")
        data = f.read()
        x = json.loads(data)
        f.close()
        ret = EmpiricalAntenna(x["antenna_num"])
        ret.points = x["points"]
        ret.values = x["values"]
        ret.sv = x["sv"]
        ret.distances = x["distances"]
        ret.recalculate = True
        return ret

    @classmethod
    def from_data(self, antenna_num, el, az, gain, dates, svs):
        ret = EmpiricalAntenna(antenna_num)
        print("Ant # {}. Data has {} entries".format(antenna_num, len(el)))

        points = []
        sv = []
        values = []
        times = []
        for _el, _az, _gain, _date, _sv in zip(el, az, gain, dates, svs):
            if ~np.isnan(_gain):
                points.append((_el, _az))
                values.append(_gain)
                times.append(_date)
                sv.append(_sv)
        ret.times = times
        ret.points = points
        ret.values = values
        ret.sv = sv

        ret.recalculate = True
        return ret

    @classmethod
    def from_db(self, antenna_num, db_file=None, table="gps_signals"):
        ret = EmpiricalAntenna(antenna_num)
        conn, sql = db_connect(db_file, table=table)
        c = conn.cursor()
        c.execute(
            sql(
                "SELECT el, az, correlation, date, sv, distance FROM "
                + table
                + " WHERE (antenna=%(ph)s)"
            ),
            (antenna_num,),
        )
        # c.execute(sql("SELECT el, az, correlation, date FROM gps_signals WHERE (antenna=%(ph)s) AND el>15 AND correlation<6 AND date<%(ph)s"), (antenna_num, "2013-11-25"))
        # c.execute(sql("SELECT el, az, correlation, date FROM gps_signals WHERE (antenna=%(ph)s) AND date<%(ph)s"), (antenna_num, "2013-11-25"))
        pval = c.fetchall()
        conn.close()
        print("Database has {} entries".format(len(pval)))
        points = []
        sv = []
        values = []
        times = []
        distances = []
        for p in pval:
            if ~np.isnan(p[2]):
                points.append((p[0], p[1]))
                values.append(p[2])
                times.append(p[3])
                sv.append(p[4])
                distances.append(p[5])
        ret.times = times
        ret.points = points
        ret.values = values
        ret.sv = sv
        ret.distances = distances
        ret.recalculate = True
        return ret



import unittest

from tart.simulation.util import *
import numpy as np


class TestUtil(unittest.TestCase):
    def test_rad2deg(self):
        self.assertEqual(Util.rad2deg(0.0), 0.0)
        self.assertEqual(Util.rad2deg(2 * Util.PI), 360.0)
        self.assertEqual(Util.rad2deg(Util.PI), 180.0)

    def test_rem(self):
        for k in range(1000):
            x = np.random.uniform(-1000, 1000)
            y = np.random.uniform(-100, 100)
            self.assertAlmostEqual(Util.rem(x, y), x % y, 9)

    def test_rem2pi(self):
        self.assertEqual(Util.rem2pi(Util.PI), Util.PI)
        self.assertEqual(Util.rem2pi(Util.PI + Util.PI2), Util.PI)
        self.assertEqual(Util.rem2pi(Util.PI - Util.PI2), Util.PI)
        self.assertEqual(Util.rem2pi(Util.PI / 10 - Util.PI2), Util.PI / 10)

    def test_idiv(self):
        for k in range(1000):
            x = np.random.uniform(-1000, 1000)
            y = np.random.uniform(-100, 100)
            self.assertEqual(Util.idiv(x, y), x // y)



'''
    Database Connection Utilities for TART
    
    Author: Tim Molteno 2014-2019 tim@elec.ac.nz
    Copyright GPLv3
'''

import sqlite3
import os

try:
    import psycopg2
except Exception as e:
    pass


def db_remove_antenna_measurements(conn, sql, utc_date, antenna, table='gps_data'):
    '''
        Delete a measurement from the database
    '''
    c = conn.cursor()
    c.execute(sql("DELETE FROM "+table+" WHERE (date=%(ph)s) AND (antenna=%(ph)s)"), (utc_date, antenna))
    conn.commit()

def db_insert_row(cursor, sql, utc_date, antenna, sv, el, az, correlation, distance, table='gps_data'):
    '''
        Insert a row into the database
    '''
    cursor.execute(sql("INSERT INTO "+table+" VALUES (%(ph)s,%(ph)s,%(ph)s,%(ph)s,%(ph)s,%(ph)s,%(ph)s)"),\
        (utc_date, sv, antenna, el, az, correlation, distance))


def db_connect(dbfile=None, table='gps_data'):
    if (dbfile != None):
        conn = sqlite3.connect(dbfile, timeout=60)
        paramstyle = sqlite3.paramstyle
    else:
        db_host= os.environ['POSTGRES_HOST']
        db_password = os.environ['POSTGRES_PASSWORD']
        db_user = os.environ['POSTGRES_USER']
        conn_parameter = "host='{}' dbname=tart2 user={} password='{}'".format(db_host, db_user, db_password)

        conn = psycopg2.connect(conn_parameter)
        paramstyle = psycopg2.paramstyle

    # Create table if necessary
    c = conn.cursor()
    c.execute("CREATE TABLE IF NOT EXISTS "+table+" (date timestamp, sv INTEGER, antenna INTEGER, el REAL, az REAL, correlation REAL, distance REAL)")
    conn.commit()
    def sql_(cmd, paramstyle):
        if paramstyle == 'qmark':
            ph = "?"
        elif paramstyle == 'pyformat':
            ph = "%s"
        else:
            raise Exception("Unexpected paramstyle: %s" % paramstyle)
        return cmd % {"ph": ph}
    sql = lambda cmd: sql_(cmd, paramstyle)
    return conn, sql



"""
Generate UVFITS file from visibility object.
"""

from astropy.io import fits
import numpy as np

from tart.imaging import tart_util
from tart.imaging import radio_source
from tart.imaging import location

from tart.simulation import antennas

from tart.util import angle
from tart.util import skyloc
from tart.util import constants


def encode_baseline(a1, a2):
    """Encode the baseline. Use the miriad convention to handle more than 255 antennas (up to 2048)."""
    if a1 > a2:
        te = a2
        a2 = a1
        a1 = te

    if a2 > 255:
        return (
            a1 * 2048 + a2 + 65536
        )  # this is backwards compatible with the standard UVFITS convention
    else:
        return a1 * 256 + a2


def decode_baseline(bl):
    """Decode a baseline using same (extended) miriad format as above"""
    if bl > 65536:
        bl -= 65536
        a2 = bl % 2048
        a1 = (bl - a2) / 2048
        return a1, a2
    else:
        a2 = bl % 256
        a1 = (bl - a2) / 256
        return a1, a2


class UVFitsGenerator(object):
    """ Generate a UVFITS file for a target source from an array of visibilities."""

    def __init__(self, cal_vis_list, phase_center):  # FIXME
        # self.phase_center_elevation = phase_center_elevation
        # self.phase_center_azimuth = phase_center_azimuth
        vt = cal_vis_list[0]

        c = vt.get_config()
        ant_p = np.asarray(c.get_antenna_positions())

        self.config = c
        self.v_array = cal_vis_list
        self.n_baselines = len(vt.get_baselines())
        self.loc = c.get_loc()  # Assume the array isn't moving

        if phase_center is None:
            ra, dec = location.get_loc(vt.get_config()).horizontal_to_equatorial(
                vt.get_timestamp(), angle.from_dms(90), angle.from_dms(0)
            )
            self.phase_center = radio_source.CosmicSource(ra, dec, 1e10)
        else:
            self.phase_center = phase_center

    def gen_vis_table(self):
        """Generate visibility table and fill payload."""

        baselines = []
        ant_pos = self.config.get_antenna_positions()
        loc = self.config.get_loc()
        for cal_vis in self.v_array:
            ra, dec = self.phase_center.radec(cal_vis.vis.timestamp)
            # v.rotate(skyloc.Skyloc(ra, dec))
            uu_a, vv_a, ww_a = cal_vis.get_all_uvw()
            bls = cal_vis.get_baselines()
            datestamp = tart_util.get_julian_date(cal_vis.vis.timestamp) - int(
                tart_util.get_julian_date(cal_vis.vis.timestamp) + 0.5
            )
            for uu, vv, ww, b in zip(uu_a, vv_a, ww_a, bls):
                baseline = {}
                [i, j] = b
                # print((np.array(a1.enu) - np.array(a0.enu)), uu, vv, ww)
                # arcane units of UVFITS require u,v,w in nanoseconds
                baseline["UU"] = uu  / constants.V_LIGHT
                baseline["VV"] = vv  / constants.V_LIGHT
                baseline["WW"] = ww  / constants.V_LIGHT
                baseline["BASELINE"] = encode_baseline(i + 1, j + 1)
                baseline["DATE"] = datestamp
                # DATE FIXME ?
                baselines.append(baseline)

        freqs = np.array([1545.0])
        pols = np.array(["+"])

        data = np.zeros(
            (len(self.v_array) * self.n_baselines, 1, 1, len(freqs), len(pols), 3)
        )

        for i, v in enumerate(self.v_array):
            for k, b in enumerate(cal_vis.get_baselines()):
                for l, _ in enumerate(freqs):
                    for j, _ in enumerate(pols):
                        vis = v.get_visibility(b[0], b[1])
                        re = vis.real
                        img = vis.imag
                        w = np.ones(1)
                        data[i * self.n_baselines + k, 0, 0, l, j, :] = [re, img, w]

        hdu = fits.GroupsHDU(
            fits.GroupData(
                data,
                parnames=["UU", "VV", "WW", "BASELINE", "DATE"],
                bitpix=-32,
                pardata=[
                    [b["UU"] for b in baselines],
                    [b["VV"] for b in baselines],
                    [b["WW"] for b in baselines],
                    [b["BASELINE"] for b in baselines],
                    [b["DATE"] for b in baselines],
                ],
            )
        )
        return hdu

    def gen_ant_table(self):
        """Generate antenna table."""
        v = self.v_array[0]
        n_ant = self.config.get_num_antenna()
        ant_pos = self.config.get_antenna_positions()
        v.antennas = []
        for i in range(n_ant):
            antenna = {}
            antenna["ANNAME"] = "ANT%i" % (i + 1)  # there is no antenna 0
            antenna["STABXYZ"] = ant_pos[i]
            antenna[
                "ORBPARM"
            ] = 0.0  # SET TO 0. - not used, because 'NUMORB' in header is    0
            antenna["NOSTA"] = i + 1  # there is no antenna 0
            antenna["MNTSTA"] = 0
            antenna["STAXOF"] = 0.0
            antenna["POLTYA"] = "X"
            antenna["POLAA"] = [0.0]
            antenna["POLCALA"] = np.array([0.0, 0.0, 0.0])
            antenna["POLTYB"] = "Y"
            antenna["POLAB"] = [0.0]
            antenna["POLCALB"] = np.array([0.0, 0.0, 0.0])
            v.antennas.append(antenna)

        c1 = fits.Column(
            name="ANNAME", format="8A", array=[a["ANNAME"] for a in v.antennas]
        )
        c2 = fits.Column(
            name="STABXYZ",
            format="3D",
            array=[a["STABXYZ"] for a in v.antennas],
            unit="METERS",
        )
        c3 = fits.Column(
            name="NOSTA", format="1J", array=[a["NOSTA"] for a in v.antennas]
        )
        c4 = fits.Column(
            name="MNTSTA", format="1J", array=[a["MNTSTA"] for a in v.antennas]
        )
        c5 = fits.Column(
            name="STAXOF",
            format="1E",
            array=[a["STAXOF"] for a in v.antennas],
            unit="METERS",
        )
        c6 = fits.Column(
            name="POLTYA", format="1A", array=[a["POLTYA"] for a in v.antennas]
        )
        c7 = fits.Column(
            name="POLAA",
            format="1E",
            array=[a["POLAA"] for a in v.antennas],
            unit="DEGREES",
        )
        c8 = fits.Column(
            name="POLCALA", format="3E", array=[a["POLCALA"] for a in v.antennas]
        )
        c9 = fits.Column(
            name="POLTYB", format="1A", array=[a["POLTYB"] for a in v.antennas]
        )
        c10 = fits.Column(
            name="POLAB",
            format="1E",
            array=[a["POLAB"] for a in v.antennas],
            unit="DEGREES",
        )
        c11 = fits.Column(
            name="POLCALB", format="3E", array=[a["POLCALB"] for a in v.antennas]
        )
        c0 = fits.Column(
            name="ORBPARM", format="1D", array=[a["ORBPARM"] for a in v.antennas]
        )

        return fits.BinTableHDU.from_columns(
            [c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, c0]
        )

    def update_vis_header(self):
        """Update visibility header."""

        v = self.v_array[0]
        params = ["UU", "VV", "WW", "BASELINE", "DATE"]
        ctypes = ["STOKES", "FREQ", "RA", "DEC"]

        self.vis_header.set("OBJECT", "Full Sky")

        # without rotation of the visibilities is has to be this:
        # ra, dec = self.loc.horizontal_to_equatorial(v.vis.timestamp, angle.from_dms(90.), angle.from_dms(0.))
        # with rotation is has to be:

        ra, dec = self.phase_center.radec(v.vis.timestamp)

        self.vis_header.set("OBSRA", ra.to_degrees())  # Antenna pointing center
        self.vis_header.set("OBSDEC", dec.to_degrees())  # Antenna pointing center

        self.vis_header.set("TELESCOP", "TART")
        self.vis_header.set("INSTRUME", "TART")
        self.vis_header.set("EPOCH", 2000.0)
        self.vis_header.set("BSCALE", 1)
        self.vis_header.set(
            "DATE-OBS",
            "%i-%02i-%02iT00:00:00.0"
            % (v.vis.timestamp.year, v.vis.timestamp.month, v.vis.timestamp.day),
        )

        for i, par in enumerate(params):
            self.vis_header.set("PTYPE%i" % (i + 1), par)
            self.vis_header.set("PSCAL%i" % (i + 1), 1.0)
            self.vis_header.set("PZERO%i" % (i + 1), 0.0)

        self.vis_header.set(
            "PZERO5", int(tart_util.get_julian_date(v.vis.timestamp) + 0.5)
        )

        self.vis_header.set("CRVAL2", 1.0)
        self.vis_header.set("CRPIX2", 1.0)
        self.vis_header.set("CDELT2", 1.0)
        self.vis_header.set("CTYPE2", "COMPLEX")

        for i, ctyp in enumerate(ctypes):
            self.vis_header.set("CTYPE%d" % (i + 3), ctyp)
            self.vis_header.set("CRVAL%d" % (i + 3), 1.0)
            self.vis_header.set("CRPIX%d" % (i + 3), 1.0)
            self.vis_header.set("CDELT%d" % (i + 3), 1.0)

        # set the input pol type.
        # for circular CRVAL3 should be -1, CDELT3 -1
        # for linear     CRVAL3 should be -5, CDELT3 -1
        # for Stokes     CRVAL3 should be    1, CDELT3    1
        # header['CRVAL3']    = -5.
        # header['CDELT3']    = -1.

        pol_typ = 1.0  # FIXME
        # pol_typ =    Obs.pol_typ
        self.vis_header.set("CRPIX3", pol_typ)
        if pol_typ < 0:
            self.vis_header.set("CDELT3", -1.0)
        else:
            self.vis_header.set("CDELT3", 1.0)

        self.vis_header.set("CRPIX4", 1.0)  # something .... n_freq/2 +1
        self.vis_header.set("CDELT4", v.vis.config.get_bandwidth())  # freq delta
        self.vis_header.set(
            "CRVAL4", v.vis.config.get_operating_frequency()
        )  # center freq

        ra, dec = self.phase_center.radec(v.vis.timestamp)

        self.vis_header.set("CRVAL5", ra.to_degrees())  # SOURCE RA    PARAMETER
        self.vis_header.set("CRVAL6", dec.to_degrees())  # SOURCE DEC PARAMETER

        self.vis_header.add_history(
            "AIPS WTSCAL =    1.0"
        )  # has to be in here for some reason
        self.vis_header.add_comment(
            "written by the TART UV FITS writer. ELEC Otago Uni"
        )

    def update_ant_header(self):
        """Update antenna header."""

        v = self.v_array[0]

        self.ant_header.set("TFORM3", "1J")  #
        self.ant_header.set("TFORM4", "1J")  #
        self.ant_header.set(
            "TFORM5", "1E"
        )  #     replace (J,E,D) with (1J,1E,1D) to match difmap requirements
        self.ant_header.set(
            "TFORM7", "1E"
        )  # # ra and dec of sun here! FIXME - move to fixed phase center in ra/dec
        self.ant_header.set("TFORM10", "1E")  #
        self.ant_header.set("TFORM12", "1D")  #

        self.ant_header.set("EXTNAME", "AIPS AN")
        self.ant_header.set("XTENSION", "BINTABLE")
        arr_loc = self.loc.get_ecef()
        self.ant_header.set("ARRAYX", arr_loc[0])  #
        self.ant_header.set(
            "ARRAYY", arr_loc[1]
        )  #    TART POSITION in ECEF coordinates
        self.ant_header.set("ARRAYZ", arr_loc[2])  #
        self.ant_header.set("FREQ", v.vis.config.get_operating_frequency())
        # /* GSTIAO is the GST at zero hours in the time system of TIMSYS (i.e. UTC) */
        # mjd = trunc(data->date[0] - 2400000.5);
        # temp = slaGmst(mjd)*180.0/M_PI;    # technically, slaGmst takes UT1, but it won't matter here.
        # fits_update_key(fptr,TDOUBLE,"GSTIA0",&temp , NULL, &status);
        self.ant_header.set("GSTIA0", self.loc.GST(v.vis.timestamp).to_degrees())
        self.ant_header.set("DEGPDY", 3.60985642988e02)  # Rotation per day

        self.ant_header.set(
            "RDATE",
            "%d-%02d-%02dT00:00:00.0"
            % (v.vis.timestamp.year, v.vis.timestamp.month, v.vis.timestamp.day),
        )
        self.ant_header.set("POLARX", 0.0)
        self.ant_header.set("POLARY", 0.0)
        self.ant_header.set("UT1UTC", 0.0)
        self.ant_header.set("DATUTC", 0.0)
        self.ant_header.set("TIMSYS", "UTC")
        self.ant_header.set("ARRNAM", "TART")
        self.ant_header.set("NUMORB", 0)
        self.ant_header.set("NOPCAL", 3)
        self.ant_header.set("FREQID", -1)
        self.ant_header.set("IATUTC", 33.0)
        self.ant_header.set("EXTVER", 1)
        # self.ant_header.set('XYZHAND','RIGHT')

    def write(self, filename):
        """Generate tables and headers and write UVFITS file."""
        vis_table = self.gen_vis_table()
        ant_table = self.gen_ant_table()

        self.vis_header = vis_table.header
        self.update_vis_header()
        self.ant_header = ant_table.header
        self.update_ant_header()

        hdulist = fits.hdu.hdulist.HDUList([vis_table, ant_table])
        hdulist.writeto(filename)



from tart.simulation.test.test_antennas import TestAntenna
from tart.simulation.test.test_antennas import TestAntennas
from tart.simulation.test.test_radio import TestMax2769B
from tart.simulation.test.test_skymodel import TestSkymodel

# from tart.simulation.test.test_simulated_visibility import TestSimulatedVisibility # FIXME broken
from tart.simulation.test.test_simulated_visibility import TestTelescope
from tart.simulation.test.test_simulation_source import TestSimulationSource
from tart.simulation.test.test_util import TestUtil



import numpy as np
import datetime

import os

try:
   import cPickle as pickle
except:
   import pickle
import math
import gzip
import h5py
import dateutil.parser

from tart.imaging import tart_util

from tart.operation import settings

from tart.util.utc import to_utc

def boolean_mean(x_bool_arr):
    '''Return mean of boolean array where -1 -> 0    and 1 -> 1'''
    ret = x_bool_arr.sum()/float(len(x_bool_arr))*2.-1
    return ret


class Observation(object):
    '''Antenna positions are going to be in meters from the array reference position.
    They will be in 3D ENU co-ordinates'''
    
    
    def __init__(self, timestamp, config, data=None, savedata=None):
        '''
            Create an observation from binary (boolean 0...1) data
            
        '''
        self.timestamp = timestamp
        self.config = config
        self.data = np.asarray(data)
        self.savedata = savedata

    def get_means(self):
        '''Calculate and return means of antenna data'''
        ret = []
        for i in range(self.config.get_num_antenna()):
            ret.append(boolean_mean(self.data[i]))
        return np.array(ret)

    def get_antenna(self, ant_num):
        if ant_num >= self.config.get_num_antenna():
            raise "Antenna %d doesn't exist" % ant_num
        return self.data[ant_num]*2-1. # Return to bipolar binary

    def get_sampling_rate(self):
        return self.config.get_sampling_frequency()    # See the Max 2769 data sheet. We operate in one of the predefined modes

    def get_julian_date(self):
        return tart_util.get_julian_date(self.timestamp)

    def get_mjd(self):
        return tart_util.get_mjd(self.timestamp)

    def save(self, filename):
        ''' Save the Observation object,
            Data is saved as one bit
        '''
        d = {}
        d['config'] = self.config.Dict
        d['timestamp'] = self.timestamp

        if self.savedata is None:
            #self.savedata = np.array((self.data+1)/2,dtype=np.uint8)
            self.savedata = np.array(self.data, dtype=np.uint8)

        t = []
        for ant in self.savedata:
            t.append(np.packbits(ant))

        d['data'] = t

        save_ptr = gzip.open(filename, 'wb')
        pickle.dump(d, save_ptr, pickle.HIGHEST_PROTOCOL)
        save_ptr.close()


    def to_hdf5(self, filename):
        ''' Save the Observation object,
            in a portable HDF5 format
            
            obs = observation.Observation(t_stmp, config, savedata=ant_data)
            obs.to_hdf5(filename)

        '''
        with h5py.File(filename, "w") as h5f:
            dt = h5py.special_dtype(vlen=bytes)
            
            conf_dset = h5f.create_dataset('config', (1,), dtype=dt)
            conf_dset[0] = self.config.to_json()

            ts_dset = h5f.create_dataset('timestamp', (1,), dtype=dt)
            ts_dset[0] = self.timestamp.isoformat()

            if self.savedata is None:
                #self.savedata = np.array((self.data+1)/2,dtype=np.uint8)
                self.savedata = np.array(self.data, dtype=np.uint8)

            t = []
            for ant in self.savedata:
                t.append(np.packbits(ant))

            h5f.create_dataset('data', data=t)

    @classmethod
    def from_hdf5(self, filename):
        ''' Load the Observation object,
            in a portable HDF5 format
        '''
        with h5py.File(filename, "r") as h5f:
            config_json = np.string_(h5f['config'][0])
            config = settings.from_json(config_json)
            timestamp = to_utc(dateutil.parser.parse(h5f['timestamp'][0]))
            
            hdf_data = h5f['data'][:]
            unipolar_data = []
            for i in hdf_data:
                unpacked_ints = np.asarray(np.unpackbits(i), dtype=np.uint8)
                unipolar_data.append(unpacked_ints)
                # this is an array of unipolar 0,1 radio signals.

        ret = Observation(timestamp=timestamp, config=config, data=unipolar_data)
        return ret

def Observation_Load(filename):
    _, file_extension = os.path.splitext(filename)

    if ('.pkl' == file_extension):
        try:
            load_data = gzip.open(filename, 'rb')
            d = pickle.load(load_data)
        except:
            print('not gzipped')
            load_data = open(filename, 'rb')
            d = pickle.load(load_data)
        finally:
            load_data.close()

        unipolar_data = []
        for i in d['data']:
            unpacked_ints = np.asarray(np.unpackbits(i), dtype=np.uint8)
            unipolar_data.append(unpacked_ints)
            # this is an array of unipolar 0,1 radio signals.

        return Observation(timestamp=d['timestamp'], config=settings.from_dict(d['config']), data=unipolar_data)
    if ('.hdf' == file_extension):
        return Observation.from_hdf5(filename)
    
    raise RuntimeError("Unknown file extension {}".format(file_extension))



import unittest
from tart.imaging.gps_time import *


class TestGpsTime(unittest.TestCase):
    def test_jump(self):
        g1 = GpsTime(2013, 9, 21, 0, 59, 3)
        g2 = GpsTime(2013, 9, 21, 1, 0, 3)
        self.assertEqual(g1.m_week, g2.m_week)
        self.assertEqual(g1.sec_of_week, g2.sec_of_week - 60.0)
        g1 = GpsTime.from_time(utc.utc_datetime(2013, 9, 21, 0, 59, 3))
        g2 = GpsTime.from_time(utc.utc_datetime(2013, 9, 21, 1, 0, 3))
        self.assertEqual(g1.m_week, g2.m_week)
        self.assertEqual(g1.sec_of_week, g2.sec_of_week - 60.0)

    def test_from_time(self):
        g1 = GpsTime.from_time(utc.utc_datetime(2013, 9, 21, 0, 59, 3))
        g2 = GpsTime(2013, 9, 21, 0, 59, 3)
        self.assertEqual(g1.m_week, g2.m_week)
        self.assertEqual(g1.sec_of_week, g2.sec_of_week)



#
# An object that represents a location in the Elevation and Azimuth, and
# handles projection into the l,m plane.
#
# Tim Molteno 2017-2019
#
import numpy as np


class ElAz(object):
    def __init__(self, el, az):
        self.el_r = np.radians(el)
        self.az_r = np.radians(az)

        # l,m are the cosines of the angle between the source vector and the u and v axes
        # respectively.
        self.l = -np.sin(self.az_r) * np.cos(self.el_r)
        self.m = np.cos(self.az_r) * np.cos(self.el_r)

    def get_lm(self):
        return self.l, self.m

    """
        Return the area of a unit-solid area of source dS
        in the l,m plane
    """

    def get_lm_area(self, dS=1.0):
        return dS * np.sqrt(1.0 - self.l * self.l - self.m * self.m)

    def get_px(self, num_bins):
        ''' Get source location in pixels from it's direction
            cosines. These should be image coordinates in an
            image num_bins x num_bins essentially used as array
            indices.
        '''
        n2 = num_bins / 2
        x_px = int(np.round(self.l * n2 + n2))
        y_px = num_bins - int(np.round(self.m * n2 + n2))
        return x_px, y_px

    def deg_to_pix(self, num_bins, deg):
        pix_per_rad = num_bins / np.pi
        d = np.radians(deg) * pix_per_rad / 2
        return d

    def get_px_window(self, num_bins, window_deg):
        ''' Get a pixel window around a source with width window_deg
            This is assumed to be an inverse FFT image with num_bins x num_bins
        '''
        x_i, y_i = self.get_px(num_bins)

        d = self.deg_to_pix(num_bins, window_deg)

        x_min = int(np.floor(x_i - d))
        x_max = int(np.ceil(x_i + d))
        y_min = int(np.floor(y_i - d))
        y_max = int(np.ceil(y_i + d))
        area = (x_max - x_min) * (y_max - y_min)
        return x_min, x_max, y_min, y_max, area

    def get_old_lm(self):
        R = 10 * np.pi * 2
        r = 90 - np.degrees(self.el_r)
        r = R * np.arctan(r / R)
        x = r * np.sin(self.az_r)
        y = r * np.cos(self.az_r)
        return [-x, y]


def from_json(source_json, el_limit=0.0, jy_limit=1e5):
    src_list = []

    for src in source_json:
        try:
            if (src["el"] > el_limit) and (src["jy"] > jy_limit):
                src_list.append(ElAz(src["el"], src["az"]))
        except:
            print(("ERROR in catalog src={}".format(src)))

    return src_list


def get_source_coordinates(source_list):
    x_list = []
    y_list = []

    for src in source_list:
        x_list.append(src.l)
        y_list.append(src.m)

    return [x_list, y_list]


if __name__ == "__main__":
    el_range = np.radians([0, 5, 30, 60, 90])
    az_range = np.radians(np.linspace(0, 360, 30, endpoint=False))

    import matplotlib.pyplot as plt

    old = []
    new = []

    for el in el_range:
        for az in az_range:
            s = ElAz(np.degrees(el), np.degrees(az))
            new.append([s.l, s.m])
            old.append(s.get_old_lm())

    old = np.array(old)
    new = np.array(new) * np.max(old)
    print((new.shape, old.shape))
    plt.figure()
    plt.scatter(old[:, 0], old[:, 1], label="old", alpha=1.0)
    plt.scatter(new[:, 0], new[:, 1], label="new", alpha=0.5)
    plt.xlabel("Elevation")
    plt.ylabel("Pixel Coordinate")
    plt.grid(True)
    plt.legend()
    plt.savefig("Old_vs_new pixel coordinates")
    plt.show()



import unittest
import datetime
import numpy as np

from tart.operation.observation import Observation, Observation_Load
from tart.operation import settings


class TestObservation(unittest.TestCase):

    def setUp(self):

        ts = datetime.datetime.utcnow()
        self.config = settings.from_file('tart/test/test_telescope_config.json')
        self.test_len = 2**8
        # generate some fake data
        self.data = [np.random.randint(0,2,self.test_len) for _ in range(self.config.get_num_antenna())]
        self.obs = Observation(timestamp=ts, config=self.config, data=self.data)


    def test_hdf5_load_save(self):
        self.obs.to_hdf5('data.hdf')

        nobs = Observation.from_hdf5('data.hdf')

        self.assertTrue((self.obs.get_antenna(1) == nobs.get_antenna(1)).all())
        self.assertEqual(self.obs.get_julian_date(), nobs.get_julian_date())
        self.assertTrue((self.data == nobs.data).all())


    #def test_str2bits(self):
        #init = '101000'
        #res = np.array([1,0,1,0,0,0])
        #self.assertTrue((res == str2bits(init)[0]).all())

    #def test_bit2int(self):
        #init = np.array([1,0,1,0,0,0,1,0,1,0,0,0])
        #res = np.array([ 2**7+2**5+2**1, 2**3])
        #self.assertTrue((bit2int(init) == res).all())

    #def test_bits2str(self):
        #init = np.array([1,0,1,0,0,0,1,0,1,0,0,0])
        #res = np.array([ 2**7+2**5+2**1, 2**3])
        #self.assertTrue((bit2int(init) == res).all())

    #def test_int2bin_str_with_n_leading_zeros(self):
        #self.assertTrue((int2bin_str_with_n_leading_zeros(0, 5) == '101'))
        #self.assertTrue((int2bin_str_with_n_leading_zeros(1, 5) == '101'))
        #self.assertTrue((int2bin_str_with_n_leading_zeros(2, 5) == '101'))
        #self.assertTrue((int2bin_str_with_n_leading_zeros(3, 5) == '101'))
        #self.assertTrue((int2bin_str_with_n_leading_zeros(6, 5) == '000101'))
        #self.assertTrue((int2bin_str_with_n_leading_zeros(8, 5) == '00000101'))

    #def test_int2bit(self):
        #init = np.array([244, 0, 1])
        #res1 = np.array([1, 1, 1, 1, 0, 1, 0, 0,    0, 0, 0, 0, 0, 0, 0, 0,    1])
        #res2 = np.array([1, 1, 1, 1, 0, 1, 0, 0,    0, 0, 0, 0, 0, 0, 0, 0,    0, 0, 1])
        #res3 = np.array([1, 1, 1, 1, 0, 1, 0, 0,    0, 0, 0, 0, 0, 0, 0, 0,    0, 0, 0, 0, 1])
        #res4 = np.array([1, 1, 1, 1, 0, 1, 0, 0,    0, 0, 0, 0, 0, 0, 0, 0,    0, 0, 0, 0, 0, 0, 0, 1])
        #l1=2*8 + 1
        #l2=2*8 + 3
        #l3=2*8 + 5
        #l4=3*8
        #self.assertTrue((int2bit(init, l1) == res1).all())
        #self.assertTrue((int2bit(init, l2) == res2).all())
        #self.assertTrue((int2bit(init, l3) == res3).all())
        #self.assertTrue((int2bit(init, l4) == res4).all())

    #def test_conversions(self):
        ## go twice through all conversions.
        #bitseq = np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0])
        #leng = len(bitseq)

        #intseq = bit2int(bitseq)
        #bitseq_f = int2bit(intseq, leng)
        #intseq_f = bit2int(bitseq_f)
        #bitseq_ff = int2bit(intseq_f, leng)
        #intseq_ff = bit2int(bitseq_ff)

        #self.assertTrue((bitseq == bitseq_f).all())
        #self.assertTrue((intseq == intseq_f).all())

        #self.assertTrue((bitseq == bitseq_ff).all())
        #self.assertTrue((intseq == intseq_ff).all())



# -*- coding: utf-8 -*-

# This module calculates the length of sample required from each
# astronomical radio noise source given parameters including
# the phase delays required

# An example to aid understanding;
# -3 -2 -1 0 1 2 3 4 5 ... E E+1 E+2 E+3	A sampled radio source
#     ^						0th pos phase leads 2
#              ^				Oth pos phase lags 2

# Max lag represents the position of the zero value
# Phase Lead is positive number, Phase Lag is a negative number
import numpy as np


def sample_size(ref_freq, freq_mult, sample_duration, sample_delay):

    # Calculate the required sample size excluding offsets
    base_size = int(ref_freq * freq_mult * sample_duration)

    # Calculate max and min offsets for each signal
    extrema = np.zeros((2, sample_delay.shape[0]))  # To hold extreme offsets
    extrema[0] = sample_delay.max(axis=1)  # Max values
    extrema[1] = sample_delay.min(axis=1)  # Min values

    # Calculate the sample size given the offsets
    sizes = np.ones(extrema.shape[1])  # blank array to hold sample sizes
    for sig in range(0, extrema.shape[1]):
        sizes[sig] = base_size + extrema[0, sig] - extrema[1, sig]  # Row 1 -ve vals

    max_size = sizes.max(0)

    # The maximum values indicate
    zero_off = extrema[0]

    return zero_off, max_size, base_size


if __name__ == "__main__":

    ref_freq = 16.368e6
    freq_mult = 256
    sample_duration = 1e-3

    sample_delay = np.array(
        [
            [-49, -39, -29, -19, -9, 0, 9, 19, 29, 39, 49],
            [-34, -27, -20, -13, -6, 0, 6, 13, 20, 27, 34],
            [-18, -14, -10, -7, -3, 0, 3, 7, 10, 14, 18],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [18, 14, 10, 7, 3, 0, -3, -7, -10, -14, -18],
            [34, 27, 20, 13, 6, 0, -6, -13, -20, -27, -34],
            [49, 39, 29, 19, 9, 0, -9, -19, -29, -39, -49],
        ]
    )

    zero_off, max_size, base_size = sample_size(
        ref_freq, freq_mult, sample_duration, sample_delay
    )
    print("Jump this many places from left to find phase zero position for each sample")
    print(zero_off)
    print("The base size is ", base_size)
    print("Length of sample required to accommodate all phase changes is ", max_size)



import os
import jsonrpclib
import datetime
import numpy as np

from tart.util.singleton import Singleton
from tart.util import utc

from tart.imaging import gps_time
from tart.imaging import ephemeris
from tart.imaging import sp3_interpolator


@Singleton
class EphemeridesProxy(object):
    def __init__(self):

        if "EPHEMERIS_SERVER_HOST" in os.environ:
            server_host = os.environ["EPHEMERIS_SERVER_HOST"]
        else:
            server_host = "localhost"

        self.server = jsonrpclib.Server("http://%s:8876/rpc/gps" % server_host)
        self.cache = {}
        self.sp3_cache = {}

    def get_date_hash(self, utc_date, sv):
        # Make a hash
        gpst = gps_time.GpsTime.from_time(utc_date)
        cache_hash = "%02d%04d%02d%02d%02d-%4d" % (
            sv,
            utc_date.year,
            utc_date.month,
            utc_date.day,
            utc_date.hour,
            gpst.m_week,
        )
        return cache_hash

    def get_sp3_hash(self, utc_date):
        # Make a hash
        gpst = gps_time.GpsTime.from_time(utc_date)
        cache_hash = "%04d%02d%02d-%4d" % (
            utc_date.year,
            utc_date.month,
            utc_date.day,
            gpst.m_week,
        )
        return cache_hash

    def get_ephemeris(self, utc_date, sv):
        h = self.get_date_hash(utc_date, sv)
        print("hash({}".format(h))
        try:
            eph = self.cache[h]
        except KeyError:
            eph_hash = self.server.get_ephemeris(utc_date.isoformat(), sv)
            eph = ephemeris.Ephemeris(eph_hash)
            self.cache[h] = eph
            print(("Cache miss %s, %d" % (utc_date, sv)))
        return eph

    def get_sp3_interpolator(self, utc_date):
        h = self.get_sp3_hash(utc_date)
        try:
            sp3 = self.sp3_cache[h]
        except KeyError:
            gpst = gps_time.GpsTime.from_time(utc_date)
            pts = self.server.get_interp_points(utc_date.isoformat())
            sp3 = sp3_interpolator.Sp3Interpolator(gpst, pts)
            self.sp3_cache[h] = sp3
            print(("sp3 Cache miss %s" % (utc_date)))
        return sp3

    def get_sv_position(self, utc_date, sv):
        gpst = gps_time.GpsTime.from_time(utc_date)
        eph = self.get_ephemeris(utc_date, sv)
        pos = eph.get_sv_position(gpst)
        print("get_sv_position({}, {}, {}) -> {}".format(utc_date, gpst, sv, pos))
        return np.array(pos)

    def get_sv_position_sp3(self, utc_date, sv):
        gpst = gps_time.GpsTime.from_time(utc_date)
        sp3 = self.get_sp3_interpolator(utc_date)
        pos = sp3.get_sv_position(gpst, sv)
        return np.array(pos)

    def get_sv_velocity(self, utc_date, sv):
        gpst = gps_time.GpsTime.from_time(utc_date)
        eph = self.get_ephemeris(utc_date, sv)
        return eph.get_velocity(gpst.sow())

    def get_sv_positions(self, utc_date):
        gpst = gps_time.GpsTime.from_time(utc_date)
        ret = []
        for sv in range(1, 32):
            try:
                eph = self.get_ephemeris(utc_date, sv)
                pos = eph.get_sv_position(gpst)
                ret.append([sv, pos])
            except jsonrpclib.ProtocolError:
                pass
        return ret

    def get_remote_position(self, utc_date, sv):
        return self.server.get_sv_position_sp3(utc_date.isoformat(), sv)



"""Copyright (C) Max Scheel 2013. All rights reserved"""

from tart.util import angle

# from tart.imaging import sun
from tart.imaging import radio_source
from tart.imaging import gps_satellite
from tart.imaging import location

from tart.simulation import simulation_source

import numpy as np
import healpy as hp
import matplotlib.pyplot as plt

import requests
import json
import datetime


def get_L1_srcs(ts=None):
    """
    In order to reduce the number of requests to the catalog it is adviseable to establish a local nginx proxy:
    apt-get install nginx
    vim /etc/nginx/sites-available/default

    proxy_cache_path /data/nginx/cache levels=1:2 keys_zone=my_cache:10m max_size=10g;
    server {
    ...
          location /catalog/ {
                proxy_cache my_cache;
                proxy_cache_methods GET HEAD POST;
                proxy_ignore_headers Cache-Control;
                proxy_cache_valid any 30m;
                proxy_pass https://tart.elec.ac.nz/catalog/;
          }
    """
    # catalog_server = "https://tart.elec.ac.nz/catalog"

    catalog_server = "http://localhost/catalog"
    if ts is None:
        ts = datetime.datetime.utcnow()
        r = requests.get(
            "{}/catalog?lat=-45.851820&lon=170.545448&date={}".format(
                catalog_server, ts.isoformat()
            )
        )
    else:
        r = requests.get(
            "{}/catalog?lat=-45.851820&lon=170.545448&date={}".format(
                catalog_server, ts.isoformat()
            )
        )
    return np.array(json.loads(r.text))


class Skymodel(object):
    """Ensemble of sources and their visibilities"""

    def __init__(
        self,
        n_sources,
        location,
        sun_str=2.0e5,
        sat_str=5.01e6,
        gps=True,
        l1_catalog=False,
        l1_elevation_threshold=30,
        l1_allowed=[],
        thesun=False,
        known_cosmic=True,
    ):
        self.n_sources = n_sources
        self.source_list = []
        self.known_objects = []
        self.gps_ants = []
        self.location = location

        self.l1_catalog = l1_catalog
        self.l1_elevation_threshold = l1_elevation_threshold
        self.l1_allowed = l1_allowed

        self.sun_str = sun_str
        self.sat_str = sat_str
        self.gps = gps
        self.thesun = thesun
        self.known_cosmic = known_cosmic

        self.el_threshold = 0

        if self.thesun:
            self.add_src(sun.Sun(jy=sun_str))
        if self.gps:
            for i in range(32):
                sv = gps_satellite.GpsSatellite(
                    i + 1, location=self.location, jy=sat_str
                )
                self.add_src(sv)
        # if self.known_cosmic:
        #  for src in radio_source.BrightSources:
        #    self.add_src(src)
        # for _ in range(self.n_sources):
        # ra = angle.from_dms(np.random.uniform(0., 360.))
        # dec = angle.asin(np.random.uniform(-1., 1))
        # cs = radio_source.CosmicSource(ra, dec)
        #  self.add_src(cs)

    def gen_beam(
        self, utc_date_init, utc_date_obs, config, radio, az_deg=20.0, el_deg=80.0
    ):
        """ Generate point source with constant at RA and DEC according to given az and el at time utc_date_init"""
        sources = []
        ra, dec = location.get_loc(config).horizontal_to_equatorial(
            utc_date_init, angle.from_dms(el_deg), angle.from_dms(az_deg)
        )
        src = radio_source.CosmicSource(ra, dec)
        el, az = src.to_horizontal(location.get_loc(config), utc_date_obs)
        sources.append(
            simulation_source.SimulationSource(
                amplitude=1.0,
                azimuth=az,
                elevation=el,
                sample_duration=radio.n_samples / radio.ref_freq,
            )
        )
        return sources

    def get_cum_src_flux(self, utc_date):
        """Return cumulative flux as list over all sources"""
        cumulativ_src_flux = []
        for k in range(len(self.known_objects) + 1):
            cumulativ_src_flux.append(
                np.array([src.jansky(utc_date) for src in self.known_objects[:k]]).sum()
            )
        return cumulativ_src_flux

    def get_int_src_flux(self, utc_date):
        """Return cumulative flux"""
        return self.get_cum_src_flux(utc_date)[-1]

    def gen_photons_per_src(self, utc_date, radio, config, n_samp=1):
        """ Generate n_samp photons per source"""
        sources = []
        # for src in self.known_objects:
        if self.l1_catalog:
            for src in get_L1_srcs(utc_date):
                if src["el"] > self.l1_elevation_threshold:
                    l = len(self.l1_allowed)
                    if (l == 0) or (src["name"] in self.l1_allowed["names"]):
                        if l != 0:
                            s_str = self.l1_allowed["strength_log10"][
                                self.l1_allowed["names"].index(src["name"])
                            ]
                            # print(s_str, type(s_str))
                            amplitude = np.power(10, s_str)
                        else:
                            amplitude = 1.0
                        print(src["name"], src["az"], src["el"])
                        sources.append(
                            simulation_source.SimulationSource(
                                r=src["r"],
                                amplitude=amplitude,
                                azimuth=angle.from_dms(src["az"]),
                                elevation=angle.from_dms(src["el"]),
                                sample_duration=radio.n_samples / radio.ref_freq,
                            )
                        )
                        # print('here')

        for src in self.known_objects:
            print("extra", src)
            # for src in self.get_src_objects(location.get_loc(config), utc_date):
            ra, declination = src.radec(utc_date)
            dx, dy = np.random.multivariate_normal(
                [0.0, 0.0], np.identity(2) * np.power(src.width, 2.0), n_samp
            ).T
            for j in range(n_samp):
                el, az = location.get_loc(config).equatorial_to_horizontal(
                    utc_date,
                    ra + angle.from_dms(dx[j]),
                    declination + angle.from_dms(dy[j]),
                )
                sources.append(
                    simulation_source.SimulationSource(
                        r=src.r,
                        amplitude=src.jansky(utc_date)
                        / self.get_int_src_flux(utc_date)
                        * 1.0
                        / n_samp,
                        azimuth=az,
                        elevation=el,
                        sample_duration=radio.n_samples / radio.ref_freq,
                    )
                )
        print(len(sources))
        return sources

    def gen_n_photons(self, config, utc_date, radio, n=10):
        """ Generate a total of n photons. Sources with more jansky will contribute more photons"""
        cumulativ_src_flux = self.get_cum_src_flux(utc_date)
        int_src_flux = cumulativ_src_flux[-1]
        rel_noise_flux = 0.0
        tot_flux = int_src_flux * (1.0 + rel_noise_flux)
        src_identifier = np.random.uniform(0.0, tot_flux, n)
        hi, _ = np.histogram(src_identifier, bins=cumulativ_src_flux)
        ret = []
        for src_num, count in enumerate(hi):
            src = self.known_objects[src_num]
            # Assume the sky is flat.
            # this will also cause problems at problems at boundaries of dec.
            dx, dy = np.random.multivariate_normal(
                [0.0, 0.0], np.identity(2) * np.power(src.width, 2.0), count
            ).T
            ra, declination = src.radec(utc_date)
            for j in range(count):
                el, az = location.get_loc(config).equatorial_to_horizontal(
                    utc_date,
                    ra + angle.from_dms(dx[j]),
                    declination + angle.from_dms(dy[j]),
                )
                ret.append(
                    simulation_source.SimulationSource(
                        amplitude=1.0 / n,
                        azimuth=az,
                        elevation=el,
                        sample_duration=radio.n_samples / radio.ref_freq,
                    )
                )
        return ret

    def set_el_threshold(self, threshold=0):
        self.el_threshold = threshold

    def get_src_objects(self, location, utc_date, threshold=None, debug=False):
        """optional argument threshold defines elevation threshold in deg. 0 = horizon. 90 = zenith"""
        if threshold is None:
            threshold = self.el_threshold
        ret = []
        for src in self.known_objects:
            try:
                el, az = src.to_horizontal(location=location, utc_date=utc_date)
                if el.to_degrees() > threshold:
                    ret.append(src)
            except:
                print("No position for", src)
                self.del_src(src)
        return ret

    def add_src(self, src):
        """Add source to known_objects"""
        self.known_objects.append(src)

    def del_src(self, src):
        """Remove source from known_objects"""
        self.known_objects.remove(src)

    def get_state_vector(self):
        """Return state vector"""
        state_vector = np.zeros(self.n_sources * 4)
        for i, source in enumerate(self.source_list):
            state_vector[i + (0 * self.n_sources)] = source.skyloc.ra.to_degrees()
            if source.skyloc.dec.to_degrees() > 90.0:
                state_vector[i + (1 * self.n_sources)] = (
                    source.skyloc.dec.to_degrees() - 360.0
                )
            else:
                state_vector[i + (1 * self.n_sources)] = source.skyloc.dec.to_degrees()
            state_vector[i + (2 * self.n_sources)] = source.flux
            state_vector[i + (3 * self.n_sources)] = source.width
        return [
            self.sun_str,
            self.sat_str,
            self.gps,
            self.thesun,
            self.known_cosmic,
            self.location,
            state_vector,
        ]

    def get_src_positions(self, location, utc_date):
        """ Save way of getting lists of el and az for all known_objects"""
        l_el = []
        l_az = []
        l_name = []
        for src in self.known_objects:
            try:
                el, az = src.to_horizontal(location, utc_date)
                if el.to_rad() > 0:
                    l_el.append(el.to_rad())
                    l_az.append(az.to_rad())
                    l_name.append(str(src))
                    # l_name.append('%s %2.2f %1.1e' % (str(src), el.to_degrees(), src.jansky(utc_date)))
            except:
                print("no position", src)
        return l_el, l_az, l_name

    def true_image(self, location, utc_date, nside):
        """Plot current sky."""

        l_el, l_az, l_name = self.get_src_positions(location, utc_date)

        m = np.zeros(hp.nside2npix(nside)) * hp.UNSEEN
        th = -np.array(l_el) + np.pi / 2.0
        l_phi = -np.array(l_az)
        pix = hp.pixelfunc.ang2pix(nside, th, l_phi)
        m[pix] = 1  # [src.jansky(utc_date) for src in self.known_objects]
        m = hp.sphtfunc.smoothing(m, fwhm=3.0 * np.pi / 180.0)
        # plt.figure()
        # hp.visufunc.orthview(m, title="", rot=(0,90,0), )
        return m

    def true_image_overlay(self, location, utc_date):
        """Plot current sky.
        Theta is colatitude and measured from North pole. 0 .. pi
         (straight up) |  el:  pi/2   | theta 0
         (horizon)     |  el:  0      | theta pi/2
        Th = pi/2 - el

        flip : {'astro', 'geo''}, optional
        Defines the convention of projection : 'astro'' (default, east towards left, west towards right) or 'geo' (east towards right, west towards left)
        """
        l_el, l_az, l_name = self.get_src_positions(location, utc_date)

        th = np.pi / 2.0 - np.array(l_el)
        l_phi = -np.array(l_az)
        # _ = [hp.projtext(i, j, n, rot=(0,90,0), color='black', weight='light', ha='left', va='center') for i, j, n in zip(th, l_phi, l_name)]
        _ = [
            hp.projtext(
                i,
                j,
                n,
                rot=(0, 90, 0),
                color="gray",
                alpha=0.8,
                weight="bold",
                ha="center",
                va="bottom",
            )
            for i, j, n in zip(th, l_phi, l_name)
        ]
        _ = [
            hp.projscatter(i, j, rot=(0, 90, 0), color="black", alpha=1.0, s=25)
            for i, j, n in zip(th, l_phi, l_name)
        ]
        _ = [
            hp.projscatter(i, j, rot=(0, 90, 0), color="white", alpha=1.0, s=10)
            for i, j, n in zip(th, l_phi, l_name)
        ]

        hp.projtext(np.pi / 2, 0.0, "N", rot=(0, 90, 0), va="top", ha="center")
        hp.projtext(np.pi / 2, -np.pi / 2.0, "E", rot=(0, 90, 0), va="center")
        hp.projtext(np.pi / 2, -np.pi, "S", rot=(0, 90, 0), ha="center")
        hp.projtext(
            np.pi / 2, -np.pi * 3.0 / 2.0, "W", rot=(0, 90, 0), ha="right", va="center"
        )

    def true_FFT_overlay(self, utc_date, nw, num_bin):
        """Plot current sky.
        Theta is colatitude and measured from North pole. 0 .. pi
         (straight up) |  el:  pi/2   | theta 0
         (horizon)     |  el:  0      | theta pi/2
        Th = pi/2 - el

        flip : {'astro', 'geo''}, optional
        Defines the convention of projection : 'astro'' (default, east towards left, west towards right) or 'geo' (east towards right, west towards left)
        """
        from tart.imaging.synthesis import get_max_ang

        max_ang = get_max_ang(nw, num_bin)
        l_el, l_az, l_name = self.get_src_positions(self.location, utc_date)
        l_el, l_az = np.array(l_el), np.array(l_az)

        r = 90.0 - np.degrees(l_el)
        x = r * np.sin(l_az)
        y = r * np.cos(l_az)
        plt.scatter(x, y, color="white", s=50)


def from_state_vector(state):
    """Generate skymodel from state vector"""
    sun_str, sat_str, gps, thesun, known_cosmic, location, state_vector = state
    psky = Skymodel(
        0,
        location=location,
        sun_str=sun_str,
        sat_str=sat_str,
        gps=gps,
        known_cosmic=known_cosmic,
    )
    psky.n_sources = len(state_vector) // 4
    psky.source_list = []

    for i in range(psky.n_sources):
        ra = angle.from_dms(state_vector[i + (0 * psky.n_sources)])
        dec = angle.from_dms(state_vector[i + (1 * psky.n_sources)])
        gs = radio_source.CosmicSource(
            ra,
            dec,
            jy=state_vector[i + (2 * psky.n_sources)],
            width=state_vector[i + (3 * psky.n_sources)],
        )
        psky.source_list.append(gs)
    return psky



"""Copyright (C) Max Scheel 2016. All rights reserved"""
import numpy as np
from multiprocessing import Pool
from tart.simulation import antennas
import time

try:
    import cPickle as pickle
except:
    import pickle


def get_vis_parallel(sky, cor, rad, ants, ant_models, config, time, mode="simp"):
    p = Pool()
    resultList = []
    ret_vis_list = []
    try:
        if type(sky) in [np.ndarray, list]:
            for s in sky:
                resultList.append(
                    p.apply_async(
                        get_vis, (s, cor, rad, ants, ant_models, config, time, mode)
                    )
                )
        if type(time) in [np.ndarray, list]:
            for t in time:
                resultList.append(
                    p.apply_async(
                        get_vis, (sky, cor, rad, ants, ant_models, config, t, mode)
                    )
                )
        p.close
        p.join
        for thread in resultList:
            vis = thread.get()
            if vis != None:
                ret_vis_list.append(vis)
        p.terminate()
    except KeyboardInterrupt:
        print("control-c pressed")
        p.terminate()
    return ret_vis_list


def get_vis(
    sky,
    cor,
    rad,
    ants,
    ant_models,
    config,
    timestamp,
    mode="simp",
    seed=None,
):
    np.random.seed(seed=seed)
    sources = sky.gen_photons_per_src(timestamp, radio=rad, config=config, n_samp=1)
    # sources = sky.gen_n_photons(config, timestamp, radio=rad, n=10)
    # print('debug: total flux',  np.array([src.jansky(timestamp) for src in self.known_objects]).sum())
    # print('debug: total amplitude', np.array([src.amplitude for src in sources]).sum())

    if mode == "full":
        timebase = np.arange(0, rad.sample_duration, 1.0 / rad.sampling_rate)
        ant_sigs_full = antennas.antennas_signal(ants, ant_models, sources, timebase)
        obs = rad.get_full_obs(ant_sigs_full, timestamp, config, timebase)
        v = cor.correlate_roll(obs)

    elif mode == "mini":
        v = antennas.antennas_simp_vis(
            ants, ant_models, sources, timestamp, config, rad.noise_level
        )
    else:
        ant_sigs_simp = antennas.antennas_simplified_signal(
            ants, ant_models, sources, rad.baseband_timebase, rad.int_freq, seed=seed
        )
        obs = rad.get_simplified_obs(ant_sigs_simp, timestamp, config=config, seed=seed)
        v = cor.correlate_roll(obs)
    return v


def get_vis_parallel_segmented(
    sky, cor, rad, ants, ant_models, config, time, mode="simp", segment_dir="temp/"
):
    vis_per_segment = 1000
    num_segments = int(np.ceil(len(sky) / float(vis_per_segment)))

    for i in range(num_segments):
        print("computing segment %i of %i" % (i, num_segments))
        start_idx = i * vis_per_segment
        end_idx = (i + 1) * vis_per_segment
        vis_seg = get_vis_parallel(
            sky[start_idx:end_idx],
            cor,
            rad,
            ants,
            ant_models,
            config,
            time,
            mode="simp",
        )
        pickle.dump(vis_seg, open(segment_dir + "segment_%010i.pkl" % i, "wb"))

    # //print('now combining segments.')
    ret = []
    # print(num_segments)
    # for i in range(num_segments):
    # print(i)
    # vis_seg = pickle.load( open( segment_dir+'segment_%010i.pkl' %i, "rb" ) )
    # ret += vis_seg
    return ret



import unittest
import math

from tart.util.angle import atan, atan2, asin, acos, from_dms

class TestAngle(unittest.TestCase):

    def test_atan(self):
        ax = atan(1.0)
        self.assertEqual(ax.to_degrees(), 45.0)
    def test_atan2(self):
        ax = atan2(10.,-10.)
        self.assertEqual(ax.to_degrees(), 135.0)
    def test_sin(self):
        ax = asin(1.0)
        self.assertEqual(ax.to_rad(), math.pi/2.)
    def test_acos(self):
        ax = acos(1.0)
        self.assertEqual(ax.to_hours(), 0.)
        ax = acos(-1.0)
        self.assertEqual(ax.to_degrees(), 180.)

    def test_hms(self):
        x = from_dms(167.7958333333)
        h,m,s = x.to_hms()
        self.assertEqual(h, 11)
        self.assertEqual(m, 11)
        self.assertAlmostEqual(s, 11)
        
    def test_dms(self):
        x = from_dms(180,20,11)
        h,m,s = x.to_dms()
        self.assertEqual(h, 180)
        self.assertEqual(m, 20)
        self.assertEqual(s, 11)



import unittest

from tart.imaging.test.test_correlator import TestCorrelator
from tart.imaging.test.test_correlator import TestHilbert
from tart.imaging.test.test_antenna_model import TestAntennaModel
from tart.imaging.test.test_correlator import TestCorrelator
from tart.imaging.test.test_correlator import TestHilbert
from tart.imaging.test.test_ephemerides_proxy import TestEphemeridesProxy
from tart.imaging.test.test_ephemeris import TestEphemeris
from tart.imaging.test.test_gps_satellite import TestGpsSatellite
from tart.imaging.test.test_gps_time import TestGpsTime
from tart.imaging.test.test_location import TestLocation
from tart.imaging.test.test_sun import TestSun
from tart.imaging.test.test_tart_util import TestTartUtil  # get rid of this!!!!
from tart.imaging.test.test_visibility import TestVisibility

if __name__ == "__main__":
    unittest.main()




class Skyloc(object):
    """
    A class representing a location in the sky in equatorial co-ordinates.
    Holds Right Ascension and Declination
    """
    def __init__(self, ra, dec):
        self.ra = ra
        self.dec = dec

    def to_horizontal(self, location, utc_time):
        """Returns the horizontal (el,az) coordinates for a given time"""
        return location.equatorial_to_horizontal(utc_time, self.ra, self.dec)

    @classmethod
    def from_equatorial(self, ra, dec):
        """Generate Skyloc object from equatorial coordinates"""
        return Skyloc(ra, dec)

    # Call this as Skyloc.from_horizontal()
    @classmethod
    def from_horizontal(self, location, utc_time, el, az):
        """Generate Skyloc object from horizontal coordinates"""
        ra, dec = location.horizontal_to_equatorial(utc_time, el, az)
        return Skyloc(ra, dec)



"""
Constants used throughout the TART

"""
import numpy as np

V_LIGHT = 2.99793e8 # Speed of light
TWO_PI = 2.0*np.pi

L1_FREQ = 1.57542e9
L1_OMEGA = 2.*np.pi*L1_FREQ
L1_WAVELENGTH = V_LIGHT / L1_FREQ



'''
    An implementation of the Nelder-Mead simplex method for minimization.
    Copyright T Molteno 2006. tim@physics.otago.ac.nz

    Some useful documents

    http:#www.cs.berkeley.edu/~jrs/4/lec/23
'''
import numpy as np

# Point is an N-dimensional array
# point = np.zeros(N)
# Simplex is N+1 points
# np.array((N+1,N))


class NelderMead(object):
    def __init__(self, f, debug=False):
         self.fun = f
         self.debug = debug
         self.function_eval_count = 0

    def func(self, x):
        self.function_eval_count += 1
        y = self.fun(x)
        if self.debug:
            print(y)
        return y

    ''' A Nelder Mead minimization algorithm
    \param func A function pointer to the function to be minimized
    \param start A starting simplex
    \param tolerance The tolerance of the solution. 
    \return The minimum
    '''
    def solve(self, start, tolerance, max_iterations=500):
        # start simplex and tolerance, optinal set maximum number of iterations
        reflection_coefficient = 1.
        contraction_coefficient = 0.5
        expansion_coefficient = 2.

        n = np.shape(start)[1]
        amoeba = start     # holds vertices of simplex
        # find the initial function values
        f = np.zeros(n+1)
        for j in range(n+1):
            f[j] = self.func(amoeba[j])


        for iteration_count in range(1, max_iterations+1):

            largest_vertex = 0# vertex with largest value
            smallest_vertex = 0# vertex with smallest value
            for j in range(n+1):
                if (f[j] > f[largest_vertex]):
                    largest_vertex = j
                if (f[j] < f[smallest_vertex]):
                    smallest_vertex = j
            # Test for convergence
            test = np.abs(f[largest_vertex] - f[smallest_vertex])/(tolerance+np.abs(f[largest_vertex]) + np.abs(f[smallest_vertex]))
            if test <= tolerance:
                print('convergence tolerance reached {}'.format(test))
                break
    
            centroid_pt = np.zeros(n)
            for m in range(n+1):
                if (m != largest_vertex):
                    centroid_pt = centroid_pt + amoeba[m]
            centroid_pt = centroid_pt / float(n)

            # reflect largest_vertex to new vertex reflect_point
            reflect_point = centroid_pt + (centroid_pt - amoeba[largest_vertex])*reflection_coefficient
            f_reflect = self.func(reflect_point)
    
            if f_reflect < f[smallest_vertex]:
                # We have the best point, now try expansion

                expansion_point = centroid_pt + (reflect_point-centroid_pt)*expansion_coefficient
                f_expand = self.func(expansion_point)
    
                if (f_expand < f_reflect):
                    if (self.debug):
                        print("Improved by reflection and expansion %f" % f_expand)
                    # expansion succeeded - replace the largest point
                    amoeba[largest_vertex] = expansion_point
                    f[largest_vertex] = f_expand
                    continue
                
            # OK expansion did not succeed (or was not tried). Check to see if reflection is OK
            if f_reflect < f[largest_vertex]:
                if (self.debug):
                    print("Improved by reflection %f" % f_reflect)
                # replace the highest with the reflection point
                amoeba[largest_vertex] = reflect_point
                f[largest_vertex] = f_reflect
                continue
            
            # Now try contraction
            contraction_point = centroid_pt + (reflect_point-centroid_pt)*contraction_coefficient
            f_contract = self.func(contraction_point)
            if f_contract < f[largest_vertex]:
                amoeba[largest_vertex] = contraction_point
                f[largest_vertex] = f_contract
                continue
            
            # finally try a shrink step
            if (self.debug):
                print("Doing a shrink step")
            half = 0.5
            center = amoeba[smallest_vertex]
            for vertex in range(n+1):
                if (vertex != smallest_vertex):
                    amoeba[vertex] = center + (amoeba[vertex]-center)*half
                    f[vertex] = self.func(amoeba[vertex])
                
        print("Nelder-Mead Complete")
        print("Iteration Count: " + str(iteration_count))
        print("Function Evaluations: " + str(self.function_eval_count))

        return amoeba[smallest_vertex]
    
if __name__ == '__main__':
    def ftest(x):
        d = np.array([0, 1.0, -1.0])
        return np.sum((x-d)**2)

    def gen_simplex(point, epsilon):
        from copy import deepcopy
        ret = []
        N = len(point)
        ret.append(point)
        for j in range(N):
            x = deepcopy(point)
            x[j] += epsilon
            ret.append(x)
        return np.array(ret)

    nm = NelderMead(ftest)
    start = gen_simplex([1000.0, 1.1, -1.0], 1)
    result = nm.solve(start, tolerance=1e-5, max_iterations=10240)

    print(result)



import datetime
import unittest

from astropy import time
from astropy.time import Time

from tart.imaging.tart_util import *
from tart.util.utc import now


class TestTartUtil(unittest.TestCase):
    def test_astropy(self):
        utcd = now()
        obstime = time.Time(utcd, scale="utc")

        jd = JulianDay(utcd)

        self.assertAlmostEqual(obstime.jd, jd, 4)



import math


class Util(object):
    PI = 3.1415926535897932
    PI2 = 6.2831853071795864
    C = 2.99793e8

    @classmethod
    def check_t(self, t):
        half_week = 302400.0
        tt = t

        if t > half_week:
            tt = t - 2 * half_week
        if t < -half_week:
            tt = t + 2 * half_week

        return tt

    @classmethod
    def idiv(self, x, y):
        """ Integer Division """
        return int(math.floor(x / y))

    @classmethod
    def rem(self, x, y):
        """ Remainder """
        return x - y * Util.idiv(x, y)

    #! \brief remainder = numerator - quotient * denominator
    #
    @classmethod
    def mod(self, x, y):
        ret = x - y * Util.idiv(x, y)
        return ret

    @classmethod
    def mod_int(self, x, y):
        ret = x % y
        if x < 0:
            ret = y - (-x) % y
        return ret

    @classmethod
    def rad2deg(self, x):
        return x * 180.0 / Util.PI

    @classmethod
    def rem2pi(self, x):
        return Util.rem(x, Util.PI2)



import unittest
import math

from astropy import time 
from astropy.time import Time 
from datetime import datetime

import tart.util.utc as utc

class TestUTC(unittest.TestCase):

    def test_basic(self):
        utcd = utc.utc_datetime(year=2009, month=7, day=22, hour=5, minute=9, second=50.3)
        apyt = time.Time('2009-07-22 05:09:50.3', format='iso', scale='utc')
        apyt2 = time.Time(utcd, scale='utc')

        utcd2 = apyt.to_datetime(utc.UTC)
        obstime = apyt2.to_datetime(utc.UTC)

        self.assertEqual(utcd2.year, utcd.year)
        self.assertEqual(utcd2.month, utcd.month)
        self.assertEqual(utcd2.day, utcd.day)
        self.assertEqual(utcd2.hour, utcd.hour)
        self.assertEqual(utcd2.minute, utcd.minute)
        self.assertEqual(utcd2.second, utcd.second)

        self.assertEqual((utcd - obstime).total_seconds(), 0.0)


    def test_now(self):
        now = utc.to_utc(datetime.utcnow())
        
        now_utc = utc.now()
        print(now, now_utc)
        self.assertAlmostEqual((now - now_utc).total_seconds(), 0.0, 3)



from multiprocessing import Pool

def parallelized_for_i_in(ilist, func):
    pool = Pool()
    result_list = []
    ret = []
    try:
        for iarg in ilist:
            result_list.append(pool.apply_async(func, (iarg,)))
        pool.close
        pool.join

        for thread in result_list:
            result = thread.get()
            if result is not None:
                ret.append(result)
        pool.terminate()
    except KeyboardInterrupt:
        print('control-c pressed')
        pool.terminate()
    return ret




from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

setup(name='tart',
    version='1.1.2b0',
    description='Transient Array Radio Telescope Imaging and Operation Library',
    long_description=readme,
    long_description_content_type="text/markdown",
    url='http://github.com/tmolteno/TART',
    author='Tim Molteno',
    author_email='tim@elec.ac.nz',
    license='GPLv3',
    install_requires=['numpy', 'matplotlib', 'healpy', 'astropy', 'h5py'],
    packages=['tart', 'tart.imaging', 'tart.simulation', 'tart.operation', 'tart.util',\
            'tart.test',  'tart.imaging.test', 'tart.simulation.test', 'tart.operation.test', 'tart.util.test'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Scientific/Engineering",
        "Topic :: Communications :: Ham Radio",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        'Programming Language :: Python :: 3',
        "Intended Audience :: Science/Research"])



# A UTC class that allows the construction of datetime objects with the timezone of UTC.
# Tim Molteno 2013-2022
#
from datetime import tzinfo, timedelta, datetime, timezone

UTC = timezone.utc

def utc_datetime(year, month, day, hour=0, minute=0, second=0.0):
    s = int(second)
    us = int((second - int(second)) * 1000000)
    return datetime(year=year, month=month, day=day, hour=hour, minute=minute, second=s, microsecond=us, tzinfo=timezone.utc)


def to_utc(dt):
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)
    return dt.astimezone(timezone.utc)

def now():
    return to_utc(datetime.utcnow())



"""
Solar Radio Source

Calculate the position in horizontal coordinates for the sun.
"""
#
# Copyright (c) Tim Molteno 2013. tim@elec.ac.nz
#
# import datetime
# import math
# from tart.util import angle


import unittest

from tart.imaging.elaz import *


class TestElaz(unittest.TestCase):

    def test_elaz_lmn_vertical(self):
        elaz = ElAz(90,0)
        elaz2 = ElAz(90,90)
        
        self.assertAlmostEqual(elaz.l, 0)
        self.assertAlmostEqual(elaz.m, 0)
        self.assertAlmostEqual(elaz2.l, 0)
        self.assertAlmostEqual(elaz2.m, 0)


    def test_elaz_lmn_horizontal(self):
        elaz = ElAz(0,0)    # Due north (straight up) 0, 1
        elaz2 = ElAz(0,90)  # Due west (-1,0)
        
        self.assertAlmostEqual(elaz.l, 0)
        self.assertAlmostEqual(elaz.m, 1)
        self.assertAlmostEqual(elaz2.l, -1)
        self.assertAlmostEqual(elaz2.m, 0)


    def test_elaz_px_window(self):
        elaz = ElAz(90,0)

        x_min, x_max, y_min, y_max, area = elaz.get_px_window(num_bins=128, window_deg=6)
        self.assertAlmostEqual(x_min, 61)
        self.assertAlmostEqual(x_max, 67)
        x_min, x_max, y_min, y_max, area = elaz.get_px_window(num_bins=128, window_deg=3)
        self.assertAlmostEqual(x_min, 62)
        self.assertAlmostEqual(x_max, 66)
        
        self.assertAlmostEqual(y_min, 62)
        self.assertAlmostEqual(y_max, 66)

    def test_elaz_px_vertical(self):
        
        elaz = ElAz(90,0)
        elaz2 = ElAz(90,90)

        x,y = elaz.get_px(num_bins=128)
        self.assertAlmostEqual(x, 64)
        self.assertAlmostEqual(y, 64)

        x,y = elaz2.get_px(num_bins=128)
        self.assertAlmostEqual(x, 64)
        self.assertAlmostEqual(y, 64)


    def test_elaz_px_horizontal(self):
        
        elaz = ElAz(0,0)    # Due north (straight up, center) 0, 1
        elaz2 = ElAz(0,90)  # Due west (-1,0)

        x,y = elaz.get_px(num_bins=128)
        self.assertAlmostEqual(x, 64)
        self.assertAlmostEqual(y, 0)

        x,y = elaz2.get_px(num_bins=128)
        self.assertAlmostEqual(x, 0)
        self.assertAlmostEqual(y, 64)



#
# Settings for the telescope. This provides an accessor object for the data stored in
# The telescope_config.json file.
#
# Tim Molteno 2013-2017 tim@elec.ac.nz
# Max Scheel 2017 max@max.ac.nz : do no derive any prober
import json
import numpy as np
from tart.util import angle
from tart.imaging import location

'''Antenna positions are in 3D ENU coordinates in meters'''
def rotate_location(array_orientation, localcoord):
    array_orientation = angle.from_dms(array_orientation)
    _e, _n, _u = localcoord
    c = array_orientation.cos()
    s = array_orientation.sin()
    e = _e*c - _n*s
    n = _e*s + _n*c
    u = _u
    return [e, n, u]

def from_dict(configdict):
    ret = Settings()
    ret.Dict = configdict
    return ret

def from_file(filename):
    f = open(filename, 'r')
    ret = f.read()
    f.close()
    return from_json(ret)

def from_json(data):
    dic = json.loads(data)
    ret = from_dict(dic)
    return ret

def from_api_json(config_json, ant_pos_json):
    ret = Settings()
    ret.Dict['num_antenna'] = config_json['num_antenna']
    ret.Dict['sampling_frequency'] = config_json['sampling_frequency']
    ret.Dict['frequency'] = config_json['operating_frequency']
    ret.Dict['bandwidth'] = config_json['bandwidth']
    ret.set_antenna_positions(ant_pos_json)
    loc = config_json['location']
    ret.set_geo(loc['lat'], loc['lon'], loc['alt'])
    return ret

class Settings:
    def __init__(self):
        self.Dict = {}

    def set_antenna_positions(self, ant_pos):
        self.Dict['antenna_positions'] = {'calibrated':ant_pos}
        assert (self.Dict['num_antenna'] == len(ant_pos))

    def load_antenna_positions(self,
                cal_ant_positions_file='calibrated_antenna_positions.json',
                design_antenna_positions_file = None):
        ant_pos = {}
        try:
            with open(cal_ant_positions_file, 'r') as fr:
                ant_pos['calibrated'] = json.loads(fr.read())
                fr.close()
        except:
                print('could not load ' + cal_ant_positions_file)
        try:
            if design_antenna_positions_file is not None:
                with open(design_antenna_positions_file, 'r') as fr:
                    ant_pos['design'] = json.loads(fr.read())
                    fr.close()
        except:
            print('could not load ' + design_antenna_positions_file)
        self.set_antenna_positions(ant_pos['calibrated'])

    def get_antenna_positions(self, key='calibrated'):
        if 'antenna_positions' in self.Dict:
            if key in self.Dict['antenna_positions']:
                return self.Dict['antenna_positions'][key]

    def to_json(self):
        from copy import deepcopy
        configdict = deepcopy(self.Dict)
        if 'antenna_positions' in configdict:
            configdict.pop('antenna_positions')
        json_str = json.dumps(configdict)
        return json_str

    def save(self, fname):
        f=open(fname, "w")
        f.write(self.to_json())
        f.close()

    def set_geo(self, lat, lon, alt):
        self.Dict['lat'] = lat
        self.Dict['lon'] = lon
        self.Dict['alt'] = alt

    def get_loc(self):
        return location.Location(angle.from_dms(self.Dict['lat']),
                                 angle.from_dms(self.Dict['lon']),
                                 self.Dict['alt'])

    def get_lat(self):
        return self.Dict['lat']

    def get_lon(self):
        return self.Dict['lon']

    def get_alt(self):
        return self.Dict['alt']

    def get_sampling_frequency(self):
        return self.Dict['sampling_frequency']

    def get_operating_frequency(self):
        return self.Dict['frequency']

    def get_bandwidth(self):
        return self.Dict['bandwidth']

    def get_name(self):
        return self.Dict['name']

    def get_num_antenna(self):
        return self.Dict['num_antenna']

    def calc_antenna_ENU(self, antenna_pos):
        ant_positions = []
        for x in antenna_pos:
            ant_positions.append(rotate_location(-self.Dict['array_orientation'], x))
        return ant_positions

    def plot_antenna_positions(self, c='blue', label=''):
        import matplotlib.pyplot as plt
        num_antennas = self.Dict['num_antenna']
        for key in self.Dict['antenna_positions']:
            ant_pos = self.Dict['antenna_positions'][key]
            labels = ['{0}'.format(i) for i in range(num_antennas)]
            ante = np.array([a[0] for a in ant_pos])
            antn = np.array([a[1] for a in ant_pos])
            plt.scatter(ante, antn, marker='o', s = 50.0, color=c, label=label)
            plt.xlim(-1.5,1.5)
            plt.ylim(-1.5,1.5)
            plt.xlabel('East [m]')
            plt.ylabel('North [m]')
            plt.tight_layout()
            for label, x, y in zip(labels, ante, antn):
                plt.annotate(label, xy=(x+0.02, y-0.03), fontsize=15)



from tart.imaging import location
from tart.imaging import sun
from tart.util import angle
from tart.imaging import tart_util

from tart.util import utc

import datetime
import numpy as np
import unittest

from astropy import coordinates as coord
from astropy import units as u
from astropy import time
from astropy.time import Time

Dunedin = location.Dunedin
Aachen = location.Aachen


class TestLocation(unittest.TestCase):
    # http://www.heavens-above.com/whattime.aspx?lat=-45.86391200&lng=170.51348452&loc=TART&alt=46.5&tz=NZST
    def setUp(self):
        self.utc_date = utc.utc_datetime(
            year=2009, month=7, day=22, hour=5, minute=9, second=50.3
        )

    def astropy_get_AltAz(self, r, el, az, loc, utc_date):
        """Get an astropy AltAz object
        r - distance from observer to object (m)
        el - elevation angle
        az - azimuth angle
        """
        obstime = time.Time(utc_date, scale="utc")
        eloc = coord.EarthLocation(
            lat=loc.latitude_deg() * u.deg,
            lon=loc.longitude_deg() * u.deg,
            height=loc.alt * u.m,
        )

        # altaz_frame = coord.AltAz(obstime=obstime, location=eloc,
        # temperature=15 * u.deg_C, pressure=1.010 * u.bar)
        altaz_frame = coord.AltAz(obstime=obstime, location=eloc)

        elaz = coord.SkyCoord(
            alt=el * u.deg, az=az * u.deg, distance=r * u.m, frame=altaz_frame
        )
        return elaz

    def astropy_horizontal_to_ECI(self, r, el, az, loc, utc_date):
        obstime = time.Time(utc_date, scale="utc")
        elaz = self.astropy_get_AltAz(r, el, az, loc, utc_date)
        gcrs = elaz.transform_to(coord.GCRS(obstime=obstime))
        return [gcrs.cartesian.x.value, gcrs.cartesian.y.value, gcrs.cartesian.z.value]

    def astropy_horizontal_to_ECEF(self, r, el, az, loc, utc_date):
        obstime = time.Time(utc_date, scale="utc")
        elaz = self.astropy_get_AltAz(r, el, az, loc, utc_date)
        itrf = elaz.transform_to(coord.ITRS)
        return [itrf.cartesian.x.value, itrf.cartesian.y.value, itrf.cartesian.z.value]

    """
    http://www.ngs.noaa.gov/cgi-bin/xyz_getxyz.prl
    =============================================================
                        Latitude                     Longitude            Ellip_Ht     Ellipsoid
    INPUT =     S455150.0832             E1703048.54427 46.5             GRS80        
    =============================================================


        X (Meters)        Y (Meters)        Z (Meters)        ELLIPSOID
        ------------- ------------- ------------- ---------

        -4388375.1765     733300.4705 -4554760.4738 GRS80
    """

    def test_ecef(self):
        x, y, z = Dunedin.get_ecef()
        self.assertAlmostEqual(x, -4388375.1765, 3)
        self.assertAlmostEqual(y, 733300.4705, 3)
        self.assertAlmostEqual(z, -4554760.4738, 3)

        x, y, z = Dunedin.get_ecef_wgs84()
        self.assertAlmostEqual(x, -4388375.1765, 3)
        self.assertAlmostEqual(y, 733300.4705, 3)
        self.assertAlmostEqual(z, -4554760.4738, 3)

    def test_XYX(self):
        x, y, z = Dunedin.get_XYZ(0.0, 0.0, 0.0)
        self.assertAlmostEqual(x, 0.0)
        self.assertAlmostEqual(y, 0.0)
        self.assertAlmostEqual(z, 0.0)
        x, y, z = Dunedin.get_XYZ(0.0, 10.0, 0.0)  # 100m north
        print(x, y, z)
        dx, dy, dz = Dunedin.get_ecef_delta_from_enu(0.0, 10.0, 0.0)
        print(dx, dy, dz)
        self.assertAlmostEqual(x, dx, 1)
        self.assertAlmostEqual(y, dy, 1)
        self.assertAlmostEqual(z, dz, 1)

    """ 
    The test vectors for this are generated by 
    http://xjubier.free.fr/en/site_pages/astronomy/coordinatesConverter.html
    """

    def test_get_ecef_delta_from_enu(self):
        # print('Testing ECEF2ENU2ECEF')
        x, y, z = Dunedin.get_ecef()
        e, n, u = Dunedin.ecef_to_enu(x, y, z)
        x2, y2, z2 = Dunedin.get_ecef_delta_from_enu(e, n, u)
        self.assertAlmostEqual(x, x2, 3)
        self.assertAlmostEqual(y, y2, 3)
        self.assertAlmostEqual(z, z2, 3)

    def test_horizontal_to_equatorial(self):
        # el_t = angle.from_dms(90.)
        # az_t = angle.from_dms(0.0)
        # ra, decl = Dunedin.horizontal_to_equatorial(self.utc_date, el_t, az_t)
        # st = Dunedin.LST(self.utc_date)
        # self.assertAlmostEqual(st.to_degrees(), ra.to_degrees(), 5)

        import ephem

        dnd = ephem.Observer()
        dnd.lon = str(Dunedin.lon.to_degrees())
        dnd.lat = str(Dunedin.lat.to_degrees())
        dnd.elevation = Dunedin.alt
        dnd.pressure = 0.0
        dnd.date = self.utc_date

        for az in np.arange(0, 360, 10):
            for el in np.arange(-90, 90, 5):
                az_i = angle.from_dms(az)
                el_i = angle.from_dms(el)

                ra, dec = Dunedin.horizontal_to_equatorial(self.utc_date, el_i, az_i)
                el_f, az_f = Dunedin.equatorial_to_horizontal(self.utc_date, ra, dec)

                # print(Dunedin)
                # print('INIT: el az:', el_i, az_i)
                # print('RA, DEC', ra, dec)
                # print('FINAL: el az:', el_f, az_f)

                self.assertTrue((el_i - el_f) < angle.from_dms(0.01))
                if abs(el) != 90:
                    self.assertTrue((az_i - az_f) < angle.from_dms(0.01))

    def test_equatorial_to_horizontal_dunedin(self):

        # http://www.ccdimages.com/Planets.aspx
        # get RA,dec for planets
        # verify RA and dec with ephem                                                [OK]
        # get el and az with ephem for dunedin today
        # calc el and az from RA and dec with OUR function

        import ephem

        dnd = ephem.Observer()
        dnd.lon = str(Dunedin.lon.to_degrees())
        dnd.lat = str(Dunedin.lat.to_degrees())
        dnd.elevation = Dunedin.alt
        dnd.pressure = 0.0
        datee = datetime.datetime.utcnow()
        dnd.date = datee

        objectlist = [
            ephem.Sun(),
            ephem.Mercury(),
            ephem.Venus(),
            ephem.Mars(),
            ephem.Jupiter(),
            ephem.Saturn(),
            ephem.Uranus(),
            ephem.Neptune(),
        ]
        for j in objectlist:
            j.compute(dnd)
            ra = angle.from_rad(j.ra.real)
            dec = angle.from_rad(j.dec.real)
            el, az = Dunedin.equatorial_to_horizontal(datee, ra, dec)
            self.assertAlmostEqual(el.to_rad(), j.alt.real, 3)
            self.assertAlmostEqual(az.to_rad(), j.az.real, 3)

    def test_equatorial_to_horizontal_aachen(self):

        # http://www.ccdimages.com/Planets.aspx
        # get RA,dec for planets
        # verify RA and dec with ephem                                                [OK]
        # get el and az with ephem for dunedin today
        # calc el and az from RA and dec with OUR function

        import ephem

        dnd = ephem.Observer()
        dnd.lon = str(Aachen.lon.to_degrees())
        dnd.lat = str(Aachen.lat.to_degrees())
        dnd.elevation = Aachen.alt
        dnd.pressure = 0.0
        datee = datetime.datetime.utcnow()

        dnd.date = datee
        # print(dnd.date)

        objectlist = [
            ephem.Sun(),
            ephem.Mercury(),
            ephem.Venus(),
            ephem.Mars(),
            ephem.Jupiter(),
            ephem.Saturn(),
            ephem.Uranus(),
            ephem.Neptune(),
        ]
        for j in objectlist:
            j.compute(dnd)
            ra = angle.from_rad(j.ra.real)
            dec = angle.from_rad(j.dec.real)
            el, az = Aachen.equatorial_to_horizontal(datee, ra, dec)
            self.assertAlmostEqual(el.to_rad(), j.alt.real, 3)
            self.assertAlmostEqual(az.to_rad(), j.az.real, 3)

    def test_horizontal_to_equatorial_astropy(self):
        from astropy import coordinates as coord
        from astropy import units as u
        from astropy import time
        from astropy.time import Time

        utc_date = datetime.datetime.utcnow()
        loc = Dunedin

        obstime = time.Time(utc_date, scale="utc")
        eloc = coord.EarthLocation(
            lat=loc.latitude_deg() * u.deg,
            lon=loc.longitude_deg() * u.deg,
            height=loc.alt * u.m,
        )
        altaz_frame = coord.AltAz(obstime=obstime, location=eloc)

        for el, az in zip(np.linspace(0, 90, 10), np.linspace(0, 259, 10)):

            elaz = coord.SkyCoord(alt=el * u.deg, az=az * u.deg, frame=altaz_frame)
            radec = elaz.transform_to(coord.ICRS)

            ra, dec = loc.horizontal_to_equatorial(
                utc_date, angle.from_dms(el), angle.from_dms(az)
            )

            self.assertAlmostEqual(
                radec.ra.degree, ra.to_degrees(), -1
            )  # TODO better agreement should be possible.
            self.assertAlmostEqual(
                radec.dec.degree, dec.to_degrees(), 1
            )  # TODO better agreement should be possible.

    def test_LST(self):
        """ the Right Ascension of the Zenith: it is exactly the same as your Local Sidereal Time. """
        st = Dunedin.LST(self.utc_date)
        check = angle.from_hours(12.53767)
        self.assertAlmostEqual(st.to_degrees(), check.to_degrees(), 2)

        utc_datetime = datetime.datetime(2013, 9, 12, 9, 10, 0)

        LST = Dunedin.LST(utc_datetime).to_hours() % 24

        lst_h = int(LST)
        lst_m = int((LST - int(LST)) * 60.0)
        lst_s = int(((LST - int(LST)) * 60 - int((LST - int(LST)) * 60)) * 60)

        self.assertAlmostEqual(lst_h, 19, 2)
        self.assertAlmostEqual(lst_m, 58, 2)
        self.assertAlmostEqual(lst_s, 13, 2)

    def test_LST_astropy(self):
        start_date = datetime.datetime(2013, 9, 12, 9, 10, 0)

        loc = Dunedin

        eloc = coord.EarthLocation(
            lat=loc.latitude_deg() * u.deg,
            lon=loc.longitude_deg() * u.deg,
            height=loc.alt * u.m,
        )

        for t in np.linspace(0, 1440, 50):
            utc_date = start_date + datetime.timedelta(seconds=t)
            LST = loc.LST(utc_date)
            h, m, s = LST.to_hms()
            print(LST.to_hours(), h, m, s)
            obstime = time.Time(utc_date, scale="utc")
            st = obstime.sidereal_time("mean", longitude=eloc.lon)
            print(st)
            self.assertAlmostEqual(h, st.hms.h, 5)
            self.assertAlmostEqual(m, st.hms.m, 5)
            self.assertTrue(abs(s - st.hms.s) < 0.9)

    def test_GST(self):
        utc_datetime = datetime.datetime(2013, 9, 12, 9, 10, 0)
        GST = Dunedin.GST(utc_datetime).to_hours() % 24
        gst_h = int(GST)
        gst_m = int((GST - int(GST)) * 60.0)
        gst_s = int(((GST - int(GST)) * 60 - int((GST - int(GST)) * 60)) * 60)

        self.assertAlmostEqual(gst_h, 8, 2)
        self.assertAlmostEqual(gst_m, 36, 2)
        self.assertAlmostEqual(gst_s, 9, 2)

    def test_GST_astropy(self):
        utc_date = datetime.datetime(2013, 9, 12, 9, 10, 0)
        GST = Dunedin.GST(utc_date).to_hours() % 24
        gst_h = int(GST)
        gst_m = int((GST - int(GST)) * 60.0)
        gst_s = int(((GST - int(GST)) * 60 - int((GST - int(GST)) * 60)) * 60)

        obstime = time.Time(utc_date, scale="utc")
        st = obstime.sidereal_time("apparent", "greenwich")
        print(st)
        self.assertAlmostEqual(gst_h, st.hms.h, 5)
        self.assertAlmostEqual(gst_m, st.hms.m, 5)
        self.assertAlmostEqual(gst_s, int(st.hms.s), 5)

    def test_julian_date(self):
        jd = tart_util.JulianDay(self.utc_date)
        self.assertAlmostEqual(jd, 2455034.71516, 5)

    def test_solar_elevation(self):
        t = utc.now()
        s = sun.Sun()
        ra, decl = s.radec(t)

        el, az = Dunedin.equatorial_to_horizontal(t, ra, decl)
        el1, az1 = s.to_horizontal(Dunedin, t)
        self.assertAlmostEqual(el.to_degrees(), el1.to_degrees(), 5)
        self.assertAlmostEqual(az.to_degrees(), az1.to_degrees(), 5)

    def test_eci_to_ecef(self):
        utc_date = utc.now()
        n_tests = 50
        x_arr = np.random.rand(n_tests) * 12345e3
        y_arr = np.random.rand(n_tests) * 23456e3
        z_arr = np.random.rand(n_tests) * 12345e3

        for x_in, y_in, z_in in zip(x_arr, y_arr, z_arr):
            ecef = location.eci_to_ecef(utc_date, x_in, y_in, z_in)

            x, y, z = location.ecef_to_eci(utc_date, ecef[0], ecef[1], ecef[2])
            self.assertAlmostEqual(x_in, x)
            self.assertAlmostEqual(y_in, y)
            self.assertAlmostEqual(z_in, z)

    def test_horizontal_to_ecef(self):
        theta = 90.0  # Straight Up
        phi = 0.0

        x, y, z = Dunedin.horizontal_to_ecef(
            0.0, angle.from_dms(theta), angle.from_dms(phi)
        )

        ecef = Dunedin.get_ecef()

        self.assertAlmostEqual(x, ecef[0])
        self.assertAlmostEqual(y, ecef[1])
        self.assertAlmostEqual(z, ecef[2])

    def test_horizontal_to_ecef_and_back(self):
        n_tests = 50

        theta = np.random.rand(n_tests) * np.pi / 2
        phi = np.random.rand(n_tests) * np.pi * 2
        r = 100.0
        for i in range(n_tests):
            x, y, z = Dunedin.horizontal_to_ecef(
                r, angle.from_rad(theta[i]), angle.from_rad(phi[i])
            )

            r2, theta2, phi2 = Dunedin.ecef_to_horizontal(x, y, z)

            self.assertAlmostEqual(theta2.to_rad(), theta[i])
            self.assertAlmostEqual(phi2.to_rad(), phi[i])

    def test_horizontal_to_ecef_vs_astropy(self):
        loc = Dunedin

        n_tests = 50
        el_arr = np.random.rand(n_tests) * np.pi / 2
        az_arr = np.random.rand(n_tests) * np.pi * 2
        r_arr = np.random.rand(n_tests) * 1e8

        utc_date = utc.now()

        for r, el, az in zip(r_arr, el_arr, az_arr):
            x, y, z = self.astropy_horizontal_to_ECEF(r, el, az, loc, utc_date)
            xi, yi, zi = Dunedin.horizontal_to_ecef(
                r, angle.from_dms(el), angle.from_dms(az)
            )

            self.assertAlmostEqual(x / xi, 1.0, 3)
            self.assertAlmostEqual(y / yi, 1.0, 3)
            self.assertAlmostEqual(z / zi, 1.0, 3)

    def test_horizontal_to_eci_and_back(self):

        n_tests = 50
        theta_arr = np.random.rand(n_tests) * np.pi / 2
        phi_arr = np.random.rand(n_tests) * np.pi * 2
        r = 1.3e7
        utc_date = utc.now()

        for theta, phi in zip(theta_arr, phi_arr):
            xi, yi, zi = Dunedin.horizontal_to_eci(
                r, angle.from_rad(theta), angle.from_rad(phi), utc_date
            )
            r2, theta2, phi2 = Dunedin.eci_to_horizontal(xi, yi, zi, utc_date)

            self.assertAlmostEqual(theta2.to_rad(), theta)
            self.assertAlmostEqual(phi2.to_rad(), phi)
            self.assertAlmostEqual(r2, r)

    def test_horizontal_to_eci_vs_astropy(self):
        utc_date = datetime.datetime.utcnow()
        loc = Dunedin

        n_tests = 50
        el_arr = np.random.rand(n_tests) * np.pi / 2
        az_arr = np.random.rand(n_tests) * np.pi * 2
        r_arr = np.random.rand(n_tests) * 1e8

        # utc_date = utc.now()

        for r, el, az in zip(r_arr, el_arr, az_arr):
            x, y, z = self.astropy_horizontal_to_ECI(r, el, az, loc, utc_date)
            xi, yi, zi = Dunedin.horizontal_to_eci(
                r, angle.from_dms(el), angle.from_dms(az), utc_date
            )
            delta = np.sqrt((x - xi) ** 2 + (y - yi) ** 2 + (z - zi) ** 2)
            print(f"loc_eci : {xi,yi,zi}, astropy: {x,y,z}")
            # TODO these are failing
            self.assertAlmostEqual(delta / r, 0.0, 3)
            self.assertAlmostEqual(x / xi, 1.0, 3)
            self.assertAlmostEqual(y / yi, 1.0, 3)
            self.assertAlmostEqual(z / zi, 1.0, 3)
